import pandas as pd
from typing import List
from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.utilities.handlers.query_utilities import (
    SELECTQueryParser,
    SELECTQueryExecutor,
)
from mindsdb.utilities import log
from mindsdb_sql_parser import ast
import zenpy

logger = log.getLogger(__name__)


class ZendeskUsersTable(APITable):
    """Zendesk Users Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the zendesk list users API

        Parameters
        ----------
        query : ast.Select
            Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Zendesk users

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(query, "users", self.get_columns())

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        subset_where_conditions = []
        api_filters = {}
        for op, arg1, arg2 in where_conditions:
            if arg1 in self.get_columns():
                if op != "=":
                    raise NotImplementedError(f"Unknown op: {op}. Only '=' is supported.")
                api_filters[arg1] = arg2
                subset_where_conditions.append([op, arg1, arg2])

        result = self.handler.zen_client.users(**api_filters)
        response = []
        if isinstance(result, zenpy.lib.generator.BaseResultGenerator):
            for user in result:
                response.append(user.to_dict())
        else:
            response.append(result.to_dict())

        df = pd.DataFrame(response, columns=self.get_columns())

        select_statement_executor = SELECTQueryExecutor(
            df, selected_columns, subset_where_conditions, order_by_conditions, result_limit
        )

        df = select_statement_executor.execute_query()

        return df

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "active",
            "alias",
            "chat_only",
            "created_at",
            "custom_role_id",
            "details",
            "email",
            "external_id",
            "id",
            "last_login_at",
            "locale",
            "locale_id",
            "moderator",
            "name",
            "notes",
            "only_private_comments",
            "organization_id",
            "phone",
            "photo",
            "restricted_agent",
            "role",
            "shared",
            "shared_agent",
            "signature",
            "suspended",
            "tags",
            "ticket_restriction",
            "time_zone",
            "two_factor_auth_enabled",
            "updated_at",
            "url",
            "verified",
            "iana_time_zone",
            "shared_phone_number",
            "role_type",
            "default_group_id",
            "report_csv",
        ]


class ZendeskTicketsTable(APITable):
    """Zendesk tickets Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the zendesk tickets API

        Parameters
        ----------
        query : ast.Select
            Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Ticket ID Data

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(query, "tickets", self.get_columns())

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        subset_where_conditions = []
        api_filters = {}
        for op, arg1, arg2 in where_conditions:
            if arg1 in self.get_columns():
                if op != "=":
                    raise NotImplementedError(f"Unknown op: {op}. Only '=' is supported.")
                api_filters[arg1] = arg2
                subset_where_conditions.append([op, arg1, arg2])

        result = self.handler.zen_client.tickets(**api_filters)
        response = []
        if isinstance(result, zenpy.lib.generator.BaseResultGenerator):
            for ticket in result:
                response.append(ticket.to_dict())
        else:
            response.append(result.to_dict())

        df = pd.DataFrame(response, columns=self.get_columns())

        select_statement_executor = SELECTQueryExecutor(
            df, selected_columns, subset_where_conditions, order_by_conditions, result_limit
        )

        df = select_statement_executor.execute_query()

        return df

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "assignee_id",
            "brand_id",
            "collaborator_ids",
            "created_at",
            "custom_fields",
            "description",
            "due_at",
            "external_id",
            "fields",
            "forum_topic_id",
            "group_id",
            "has_incidents",
            "id",
            "organization_id",
            "priority",
            "problem_id",
            "raw_subject",
            "recipient",
            "requester_id",
            "sharing_agreement_ids",
            "status",
            "subject",
            "submitter_id",
            "tags",
            "type",
            "updated_at",
            "url",
            "generated_timestamp",
            "follower_ids",
            "email_cc_ids",
            "is_public",
            "custom_status_id",
            "followup_ids",
            "ticket_form_id",
            "allow_channelback",
            "allow_attachments",
            "from_messaging_channel",
            "satisfaction_rating.assignee_id",
            "satisfaction_rating.created_at",
            "satisfaction_rating.group_id",
            "satisfaction_rating.id",
            "satisfaction_rating.requester_id",
            "satisfaction_rating.score",
            "satisfaction_rating.ticket_id",
            "satisfaction_rating.updated_at",
            "satisfaction_rating.url",
            "via.channel",
            "via.source.rel",
        ]


class ZendeskTriggersTable(APITable):
    """Zendesk Triggers Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the zendesk triggers API

        Parameters
        ----------
        query : ast.Select
            Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Trigger Data

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(query, "triggers", self.get_columns())

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        subset_where_conditions = []
        api_filters = {}
        for op, arg1, arg2 in where_conditions:
            if arg1 in self.get_columns():
                if op != "=":
                    raise NotImplementedError(f"Unknown op: {op}. Only '=' is supported.")
                api_filters[arg1] = arg2
                subset_where_conditions.append([op, arg1, arg2])

        result = self.handler.zen_client.triggers(**api_filters)
        response = []
        if isinstance(result, zenpy.lib.generator.BaseResultGenerator):
            for trigger in result:
                response.append(trigger.to_dict())
        else:
            response.append(result.to_dict())

        df = pd.DataFrame(response, columns=self.get_columns())

        select_statement_executor = SELECTQueryExecutor(
            df, selected_columns, subset_where_conditions, order_by_conditions, result_limit
        )

        df = select_statement_executor.execute_query()

        return df

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "actions",
            "active",
            "description",
            "id",
            "position",
            "title",
            "url",
            "updated_at",
            "created_at",
            "default",
            "raw_title",
            "category_id",
            "conditions.all",
            "conditions.any",
        ]


class ZendeskActivitiesTable(APITable):
    """Zendesk Activities Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the zendesk activities API

        Parameters
        ----------
        query : ast.Select
            Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Activity list Data

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(query, "activities", self.get_columns())

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        subset_where_conditions = []
        api_filters = {}
        for op, arg1, arg2 in where_conditions:
            if arg1 in self.get_columns():
                if op != "=":
                    raise NotImplementedError(f"Unknown op: {op}. Only '=' is supported.")
                api_filters[arg1] = arg2
                subset_where_conditions.append([op, arg1, arg2])

        result = self.handler.zen_client.activities(**api_filters)
        response = []
        if isinstance(result, zenpy.lib.generator.BaseResultGenerator):
            for activity in result:
                response.append(activity.to_dict())
        else:
            response.append(result.to_dict())

        df = pd.DataFrame(response, columns=self.get_columns())

        select_statement_executor = SELECTQueryExecutor(
            df, selected_columns, subset_where_conditions, order_by_conditions, result_limit
        )

        df = select_statement_executor.execute_query()

        return df

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "created_at",
            "id",
            "title",
            "updated_at",
            "url",
            "verb",
            "user_id",
            "actor_id",
            "actor.id",
            "actor.url",
            "actor.name",
            "actor.email",
            "actor.created_at",
            "actor.updated_at",
            "actor.time_zone",
            "actor.iana_time_zone",
            "actor.phone",
            "actor.shared_phone_number",
            "actor.photo",
            "actor.locale_id",
            "actor.locale",
            "actor.organization_id",
            "actor.role",
            "actor.verified",
            "actor.external_id",
            "actor.tags",
            "actor.alias",
            "actor.active",
            "actor.shared",
            "actor.shared_agent",
            "actor.last_login_at",
            "actor.two_factor_auth_enabled",
            "actor.signature",
            "actor.details",
            "actor.notes",
            "actor.role_type",
            "actor.custom_role_id",
            "actor.moderator",
            "actor.ticket_restriction",
            "actor.only_private_comments",
            "actor.restricted_agent",
            "actor.suspended",
            "actor.default_group_id",
            "actor.report_csv",
            "user.active",
            "user.alias",
            "user.chat_only",
            "user.created_at",
            "user.custom_role_id",
            "user.details",
            "user.email",
            "user.external_id",
            "user.id",
            "user.last_login_at",
            "user.locale",
            "user.locale_id",
            "user.moderator",
            "user.name",
            "user.notes",
            "user.only_private_comments",
            "user.organization_id",
            "user.phone",
            "user.photo",
            "user.restricted_agent",
            "user.role",
            "user.shared",
            "user.shared_agent",
            "user.signature",
            "user.suspended",
            "user.tags",
            "user.ticket_restriction",
            "user.time_zone",
            "user.two_factor_auth_enabled",
            "user.updated_at",
            "user.url",
            "user.verified",
            "user.iana_time_zone",
            "user.shared_phone_number",
            "user.role_type",
            "user.default_group_id",
            "user.report_csv",
            "target.active",
            "target.content_type",
            "target.created_at",
            "target.id",
            "target.method",
            "target.password",
            "target.target_url",
            "target.title",
            "target.type",
            "target.url",
            "target.username",
            "target.ticket.assignee_id",
            "target.ticket.brand_id",
            "target.ticket.collaborator_ids",
            "target.ticket.created_at",
            "target.ticket.custom_fields",
            "target.ticket.description",
            "target.ticket.due_at",
            "target.ticket.external_id",
            "target.ticket.fields",
            "target.ticket.forum_topic_id",
            "target.ticket.group_id",
            "target.ticket.has_incidents",
            "target.ticket.id",
            "target.ticket.organization_id",
            "target.ticket.priority",
            "target.ticket.problem_id",
            "target.ticket.raw_subject",
            "target.ticket.recipient",
            "target.ticket.requester_id",
            "target.ticket.satisfaction_rating",
            "target.ticket.sharing_agreement_ids",
            "target.ticket.status",
            "target.ticket.subject",
            "target.ticket.submitter_id",
            "target.ticket.tags",
            "target.ticket.type",
            "target.ticket.updated_at",
            "target.ticket.url",
            "target.ticket.via",
        ]



from typing import List, Dict, Any, Callable, Union
from datetime import datetime, timedelta, timezone
import pandas as pd

from mindsdb.integrations.libs.api_handler import APIResource
from mindsdb.integrations.utilities.sql_utils import FilterCondition, FilterOperator, SortColumn
from mindsdb.utilities import log


logger = log.getLogger(__name__)


def normalize_datetime_to_iso8601(dt_value: Union[str, datetime, None]) -> str:
    """
    Normalize various datetime formats to ISO 8601 with UTC timezone (Z suffix).

    Args:
        dt_value: Datetime value as string, datetime object, or None

    Returns:
        ISO 8601 formatted string with Z suffix (e.g., "2024-01-15T10:30:00Z")

    Raises:
        ValueError: If the datetime string cannot be parsed
    """
    if dt_value is None:
        return datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ")

    if isinstance(dt_value, datetime):
        # If naive (no timezone), assume UTC
        if dt_value.tzinfo is None:
            dt_value = dt_value.replace(tzinfo=timezone.utc)
        # Convert to UTC if not already
        elif dt_value.tzinfo != timezone.utc:
            dt_value = dt_value.astimezone(timezone.utc)
        return dt_value.strftime("%Y-%m-%dT%H:%M:%SZ")

    if isinstance(dt_value, str):
        dt_str = dt_value.strip()

        # Already in correct format with Z suffix
        if dt_str.endswith("Z") and "T" in dt_str:
            return dt_str

        # Has timezone offset like +00:00 or -05:00
        if dt_str.endswith(("00", "30")) and ("+" in dt_str[-6:] or dt_str[-6:-3] == "-"):
            try:
                dt_obj = datetime.fromisoformat(dt_str)
                return dt_obj.astimezone(timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ")
            except ValueError:
                pass

        formats_to_try = [
            "%Y-%m-%dT%H:%M:%S",  # ISO without timezone
            "%Y-%m-%d %H:%M:%S",  # Common format
            "%Y-%m-%d",  # Date only (assume start of day UTC)
            "%Y/%m/%d",  # Alternative date format
            "%d-%m-%Y",  # European date format
            "%m/%d/%Y",  # US date format
        ]

        for fmt in formats_to_try:
            try:
                dt_obj = datetime.strptime(dt_str, fmt)
                dt_obj = dt_obj.replace(tzinfo=timezone.utc)
                return dt_obj.strftime("%Y-%m-%dT%H:%M:%SZ")
            except ValueError:
                continue

        try:
            dt_obj = datetime.fromisoformat(dt_str.replace("Z", "+00:00"))
            return dt_obj.astimezone(timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ")
        except ValueError:
            raise ValueError(f"Unable to parse datetime string: {dt_str}")

    raise ValueError(f"Unsupported datetime type: {type(dt_value)}")


def paginate_api_call(
    api_call: Callable,
    result_key: str,
    limit: int = None,
    params: Dict[str, Any] = None,
    json_body: Dict[str, Any] = None,
    cursor_path: str = "records.cursor",
    cursor_param: str = "cursor",
    max_pages: int = 100,
) -> List[Dict]:
    """
    Helper function to paginate through API responses.

    Args:
        api_call: Function to call the API (should accept params and/or json)
        result_key: Key in response containing the data items
        limit: Maximum number of items to fetch
        params: Initial query parameters for GET requests
        json_body: Initial JSON body for POST requests
        cursor_path: Dot-notation path to cursor in response (e.g., "records.cursor")
        cursor_param: Parameter name for passing cursor to next request
        max_pages: Maximum number of pages to fetch (safety guard)

    Returns:
        List of items from paginated API calls
    """
    all_items = []
    seen_cursors = set()
    cursor = None
    page_count = 0

    params = params or {}
    json_body = json_body or {}

    while page_count < max_pages and (not limit or len(all_items) < limit):
        page_count += 1

        if cursor:
            if json_body:
                current_json = json_body.copy()
                current_json[cursor_param] = cursor
                response = api_call(json=current_json)
            else:
                current_params = params.copy()
                current_params[cursor_param] = cursor
                response = api_call(params=current_params)
        else:
            if json_body:
                response = api_call(json=json_body.copy())
            else:
                response = api_call(params=params.copy())

        items_batch = response.get(result_key, [])

        if not items_batch:
            break

        items_added_this_batch = 0
        for item in items_batch:
            if limit and len(all_items) >= limit:
                break

            item_str = str(sorted(item.items())) if isinstance(item, dict) else str(item)
            if item_str not in seen_cursors:
                all_items.append(item)
                seen_cursors.add(item_str)
                items_added_this_batch += 1

        if limit and len(all_items) >= limit:
            break

        next_cursor = response
        for key in cursor_path.split("."):
            next_cursor = next_cursor.get(key, {}) if isinstance(next_cursor, dict) else None
            if next_cursor is None:
                break

        if not next_cursor:
            break

        if next_cursor == cursor:
            logger.warning(f"API returned identical cursor: {cursor}. Stopping pagination.")
            break

        cursor_str = str(next_cursor)
        if cursor_str in seen_cursors:
            logger.warning(f"Detected cursor cycle at: {cursor_str}. Stopping pagination.")
            break

        seen_cursors.add(cursor_str)
        cursor = next_cursor

    if page_count >= max_pages:
        logger.warning(f"Reached maximum page limit ({max_pages}). There may be more data available.")

    return all_items


class GongCallsTable(APIResource):
    """The Gong Calls Table implementation"""

    def list(
        self,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        sort: List[SortColumn] = None,
        targets: List[str] = None,
    ) -> pd.DataFrame:
        """Pulls data from the Gong Calls API

        Returns
        -------
        pd.DataFrame
            Gong calls matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        api_params = {}
        if conditions:
            for condition in conditions:
                if condition.column == "date" and condition.op == FilterOperator.GREATER_THAN:
                    api_params["fromDateTime"] = normalize_datetime_to_iso8601(condition.value)
                    condition.applied = True
                elif condition.column == "date" and condition.op == FilterOperator.LESS_THAN:
                    api_params["toDateTime"] = normalize_datetime_to_iso8601(condition.value)
                    condition.applied = True

        try:
            all_calls = paginate_api_call(
                api_call=lambda params: self.handler.call_gong_api("/v2/calls", params=params),
                result_key="calls",
                limit=limit,
                params=api_params,
            )

            data = []
            for call in all_calls:
                started = call.get("started", "")
                date = started.split("T")[0] if started else ""

                item = {
                    "call_id": call.get("id"),
                    "title": call.get("title"),
                    "date": date,
                    "duration": call.get("duration"),
                    "recording_url": call.get("url", ""),
                    "call_type": call.get("system"),
                    "user_id": call.get("primaryUserId"),
                    "participants": ",".join([p.get("name", "") for p in call.get("participants", [])]),
                    "status": call.get("status"),
                }
                data.append(item)

            df = pd.DataFrame(data)

            if conditions:
                for condition in conditions:
                    if not condition.applied and condition.column in df.columns:
                        if condition.op == FilterOperator.EQUAL:
                            df = df[df[condition.column] == condition.value]
                            condition.applied = True

            if sort:
                for col in sort:
                    if col.column in df.columns:
                        df = df.sort_values(by=col.column, ascending=col.ascending, na_position="last")
                        col.applied = True
                        break
            return df

        except Exception as e:
            logger.error(f"Error fetching calls from Gong API: {e}")
            raise

    def get_columns(self) -> List[str]:
        """Returns the columns of the calls table"""
        return [
            "call_id",
            "title",
            "date",
            "duration",
            "recording_url",
            "call_type",
            "user_id",
            "participants",
            "status",
        ]


class GongUsersTable(APIResource):
    """The Gong Users Table implementation"""

    def list(
        self,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        sort: List[SortColumn] = None,
        targets: List[str] = None,
    ) -> pd.DataFrame:
        """Pulls data from the Gong Users API

        Returns
        -------
        pd.DataFrame
            Gong users matching the query
        """

        api_params = {}

        try:
            # Use pagination helper to fetch users
            all_users = paginate_api_call(
                api_call=lambda params: self.handler.call_gong_api("/v2/users", params=params),
                result_key="users",
                limit=limit,
                params=api_params,
            )

            # Process the limited data
            data = []
            for user in all_users:
                # Safely concatenate names - handle None values
                first_name = user.get("firstName") or ""
                last_name = user.get("lastName") or ""
                full_name = f"{first_name} {last_name}".strip()

                item = {
                    "user_id": user.get("id"),
                    "name": full_name,
                    "email": user.get("emailAddress", ""),
                    "role": user.get("title", ""),
                    "permissions": ",".join(user.get("permissions", [])),
                    "status": "active" if user.get("active", False) else "inactive",
                }
                data.append(item)

            df = pd.DataFrame(data)

            if conditions:
                for condition in conditions:
                    if condition.column in df.columns:
                        if condition.op == FilterOperator.EQUAL:
                            df = df[df[condition.column] == condition.value]
                            condition.applied = True

            if sort:
                for col in sort:
                    if col.column in df.columns:
                        df = df.sort_values(by=col.column, ascending=col.ascending, na_position="last")
                        col.applied = True
                        break

            if limit is not None:
                df = df.head(limit)

            return df

        except Exception as e:
            logger.error(f"Error fetching users from Gong API: {e}")
            raise

    def get_columns(self) -> List[str]:
        """Returns the columns of the users table"""
        return ["user_id", "name", "email", "role", "permissions", "status"]


class GongAnalyticsTable(APIResource):
    """The Gong Analytics Table implementation"""

    def list(
        self,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        sort: List[SortColumn] = None,
        targets: List[str] = None,
    ) -> pd.DataFrame:
        """Pulls data from the Gong Analytics API

        Returns
        -------
        pd.DataFrame
            Gong analytics matching the query
        """

        try:
            # Default to last 7 days if no date filters provided
            default_from = datetime.now(timezone.utc) - timedelta(days=7)
            default_to = datetime.now(timezone.utc)

            payload = {
                "filter": {
                    "fromDateTime": normalize_datetime_to_iso8601(default_from),
                    "toDateTime": normalize_datetime_to_iso8601(default_to),
                },
                "contentSelector": {
                    "exposedFields": {
                        "content": {
                            "brief": True,
                            "outline": True,
                            "highlights": True,
                            "callOutcome": True,
                            "topics": True,
                            "trackers": True,
                        },
                        "interaction": {"personInteractionStats": True, "questions": True},
                    }
                },
            }

            if conditions:
                for condition in conditions:
                    if condition.column == "date" and condition.op == FilterOperator.GREATER_THAN:
                        payload["filter"]["fromDateTime"] = normalize_datetime_to_iso8601(condition.value)
                        condition.applied = True
                    elif condition.column == "date" and condition.op == FilterOperator.LESS_THAN:
                        payload["filter"]["toDateTime"] = normalize_datetime_to_iso8601(condition.value)
                        condition.applied = True

            # Fetch calls using improved pagination helper with POST support
            calls_data = paginate_api_call(
                api_call=lambda **kwargs: self.handler.call_gong_api("/v2/calls/extensive", method="POST", **kwargs),
                result_key="calls",
                limit=limit,
                json_body=payload,
                cursor_path="records.cursor",
                cursor_param="cursor",
                max_pages=100,
            )

            # Process each call to extract analytics
            all_analytics = []
            for call in calls_data:
                # Extract analytics from extensive call data
                content = call.get("content", {})
                interaction = call.get("interaction", {})
                metadata = call.get("metaData", {})

                # Sentiment and Emotion from InteractionStats
                person_stats = interaction.get("interactionStats", [])
                sentiment_score = 0
                emotions = ""  # Initialize emotions

                if person_stats:
                    stats_dict = {stat["name"]: stat["value"] for stat in interaction.get("interactionStats", [])}
                    sentiment_score = (
                        stats_dict.get("Talk Ratio", 0)
                        + stats_dict.get("Patience", 0)
                        + min(stats_dict.get("Interactivity", 0) / 10, 1.0)
                    ) / 3
                    emotions = f"Talk:{stats_dict.get('Talk Ratio', 0)}, Patience:{stats_dict.get('Patience', 0)}, Interactivity:{stats_dict.get('Interactivity', 0)}"

                # Topics from AI analysis
                topics = content.get("topics", [])
                topic_names = [
                    topic.get("name", "")
                    for topic in topics
                    if isinstance(topic, dict) and topic.get("duration", 0) > 0
                ]

                # Key phrases from AI
                trackers = content.get("trackers", [])
                key_phrases = [tracker.get("name", "") for tracker in trackers if tracker.get("count", 0) > 0]

                # Topic scoring based on relevance - prevent division by zero
                topic_score = 0
                if topics:
                    valid_topics = [topic for topic in topics if isinstance(topic, dict)]
                    if valid_topics:
                        total_topic_duration = sum([topic.get("duration", 0) for topic in valid_topics])
                        avg_topic_duration = total_topic_duration / len(valid_topics)
                        call_duration = metadata.get("duration", 0)
                        if call_duration > 0:
                            topic_score = avg_topic_duration / call_duration

                item = {
                    "call_id": metadata.get("id"),
                    "sentiment_score": round(sentiment_score, 3),
                    "topic_score": round(topic_score, 3),
                    "key_phrases": ", ".join(key_phrases),
                    "topics": ", ".join(topic_names),
                    "emotions": emotions,
                    "confidence_score": "",
                }
                all_analytics.append(item)

            df = pd.DataFrame(all_analytics)

            # Apply non-date filtering at DataFrame level
            if conditions:
                for condition in conditions:
                    if not condition.applied and condition.column in df.columns:
                        if condition.op == FilterOperator.EQUAL:
                            df = df[df[condition.column] == condition.value]
                            condition.applied = True

            # Apply sorting at DataFrame level
            if sort:
                for col in sort:
                    if col.column in df.columns:
                        df = df.sort_values(by=col.column, ascending=col.ascending, na_position="last")
                        col.applied = True
                        break

            if limit is not None:
                df = df.head(limit)

            return df

        except Exception as e:
            logger.error(f"Error fetching analytics from Gong API: {e}")
            raise

    def get_columns(self) -> List[str]:
        """Returns the columns of the analytics table"""
        return ["call_id", "sentiment_score", "topic_score", "key_phrases", "topics", "emotions", "confidence_score"]


class GongTranscriptsTable(APIResource):
    """The Gong Transcripts Table implementation"""

    def list(
        self,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        sort: List[SortColumn] = None,
        targets: List[str] = None,
    ) -> pd.DataFrame:
        """Pulls data from the Gong Transcripts API

        Returns
        -------
        pd.DataFrame
            Gong transcripts matching the query
        """

        try:
            calls_api_params = {}

            if conditions:
                for condition in conditions:
                    if condition.column == "date" and condition.op == FilterOperator.GREATER_THAN:
                        calls_api_params["fromDateTime"] = normalize_datetime_to_iso8601(condition.value)
                        condition.applied = True
                    elif condition.column == "date" and condition.op == FilterOperator.LESS_THAN:
                        calls_api_params["toDateTime"] = normalize_datetime_to_iso8601(condition.value)
                        condition.applied = True

            # Fetch call IDs using pagination helper
            calls_fetch_limit = limit if limit else 100
            all_calls = paginate_api_call(
                api_call=lambda params: self.handler.call_gong_api("/v2/calls", params=params),
                result_key="calls",
                limit=calls_fetch_limit,
                params=calls_api_params,
            )
            all_call_ids = [call.get("id") for call in all_calls if call.get("id")]

            if not all_call_ids:
                return pd.DataFrame()
            call_ids_to_fetch = all_call_ids[:limit] if limit else all_call_ids

            # Fetch transcripts using improved pagination helper with POST support
            payload = {"filter": {"callIds": call_ids_to_fetch}}
            call_transcripts = paginate_api_call(
                api_call=lambda **kwargs: self.handler.call_gong_api("/v2/calls/transcript", method="POST", **kwargs),
                result_key="callTranscripts",
                limit=None,  # Get all transcripts for the specified calls
                json_body=payload,
                cursor_path="records.cursor",
                cursor_param="cursor",
                max_pages=50,
            )

            # Process transcripts
            all_transcript_data = []
            for call_transcript in call_transcripts:
                call_id = call_transcript.get("callId")
                transcript_segments = call_transcript.get("transcript", [])

                segment_counter = 0

                for speaker_block in transcript_segments:
                    speaker_id = speaker_block.get("speakerId")
                    sentences = speaker_block.get("sentences", [])

                    for sentence in sentences:
                        segment_counter += 1

                        item = {
                            "call_id": call_id,
                            "speaker": speaker_id,
                            "timestamp": sentence.get("start"),
                            "text": sentence.get("text"),
                            "confidence": sentence.get("confidence"),
                            "segment_id": f"{call_id}_{segment_counter}",
                        }
                        all_transcript_data.append(item)

            df = pd.DataFrame(all_transcript_data)

            if conditions:
                for condition in conditions:
                    if not condition.applied and condition.column in df.columns:
                        if condition.op == FilterOperator.EQUAL:
                            df = df[df[condition.column] == condition.value]
                            condition.applied = True
                        elif condition.op == FilterOperator.LIKE or condition.op == FilterOperator.CONTAINS:
                            if condition.column == "text":
                                df = df[df[condition.column].str.contains(condition.value, case=False, na=False)]
                                condition.applied = True

            if sort:
                for col in sort:
                    if col.column in df.columns:
                        df = df.sort_values(by=col.column, ascending=col.ascending, na_position="last")
                        col.applied = True
                        break

            if limit is not None:
                df = df.head(limit)
            return df

        except Exception as e:
            logger.error(f"Error fetching transcripts from Gong API: {e}")
            raise

    def get_columns(self) -> List[str]:
        """Returns the columns of the transcripts table"""
        return ["call_id", "speaker", "timestamp", "text", "confidence", "segment_id"]



"""
Constants and metadata for the Gong handler.
"""


def get_gong_api_info(handler_name: str) -> str:
    return f"""
        # Gong Handler Usage Guide: {handler_name}

        ## CRITICAL Performance Requirements

        **ALWAYS use date filters on analytics and (ideally) on calls/transcripts**
        - Analytics defaults to last 7 days if no date filters are provided
        - Calls/transcripts can be very large; add WHERE date ... or call_id filters
        - Example: WHERE c.date >= '2024-01-01' AND c.date < '2024-01-31'

        **ALWAYS use LIMIT with transcripts**
        - Transcripts can contain thousands of rows per call
        - Example: LIMIT 100

        **Query by call_id when fetching transcripts**
        - First get call IDs from `calls` (with date filters), then fetch transcripts for those IDs
        - Avoids massive data transfers

        ## Query Strategy by User Intent

        **"calls" / "meetings"** → query `{handler_name}.calls`

        **"sentiment" / "topics" / "what was discussed"** → query `{handler_name}.analytics` JOIN `{handler_name}.calls`
        - Analytics contains AI-generated insights about call content

        **"what did X say" / keyword search in speech** → query `{handler_name}.transcripts` JOIN `{handler_name}.calls`
        - Transcripts contain exact words spoken
        - Always filter by call_id and/or a narrow date range

        **"sales reps" / "users" / "team"** → query `{handler_name}.users`
        - Independent table, no JOIN needed

        ## Efficient JOIN Pattern

        Start with the smallest, filtered dataset first:
        ```sql
        SELECT c.title, a.sentiment_score
        FROM {handler_name}.calls c
        JOIN {handler_name}.analytics a ON c.call_id = a.call_id
        WHERE c.date >= '2024-01-01' AND c.date < '2024-02-01'
        LIMIT 50;
        ```
    """


GONG_TABLES_METADATA = {
    "calls": {
        "name": "calls",
        "type": "api_table",
        "description": "Call records from Gong with basic metadata including date, duration, participants, and status",
        "columns": [
            {"name": "call_id", "type": "str", "description": "Unique identifier for the call"},
            {"name": "title", "type": "str", "description": "Call title or subject"},
            {"name": "date", "type": "str", "description": "Call date (ISO 8601 format, YYYY-MM-DD)"},
            {"name": "duration", "type": "int", "description": "Call duration in seconds"},
            {"name": "recording_url", "type": "str", "description": "URL to the call recording (if available)"},
            {"name": "call_type", "type": "str", "description": "Call type/system classification"},
            {"name": "user_id", "type": "str", "description": "Primary user/owner of the call"},
            {"name": "participants", "type": "str", "description": "Comma-separated list of call participants"},
            {"name": "status", "type": "str", "description": "Call status (scheduled, completed, etc.)"},
        ],
        "filterable_columns": ["date", "status"],
        "api_endpoint": "/v2/calls",
        "supports_pagination": True,
    },
    "users": {
        "name": "users",
        "type": "api_table",
        "description": "User information including names, emails, roles, and permissions",
        "columns": [
            {"name": "user_id", "type": "str", "description": "Unique identifier for the user"},
            {"name": "name", "type": "str", "description": "User's full name"},
            {"name": "email", "type": "str", "description": "User's email address"},
            {"name": "role", "type": "str", "description": "User's role in the organization"},
            {"name": "permissions", "type": "str", "description": "User's permission levels"},
            {"name": "status", "type": "str", "description": "User status (active/inactive)"},
        ],
        "filterable_columns": ["email", "status"],
        "api_endpoint": "/v2/users",
        "supports_pagination": True,
    },
    "analytics": {
        "name": "analytics",
        "type": "api_table",
        "description": "Advanced call analytics including sentiment analysis, topics, key phrases, and interaction scores",
        "columns": [
            {"name": "call_id", "type": "str", "description": "Unique identifier for the call"},
            {"name": "sentiment_score", "type": "float", "description": "Overall sentiment score (0-1)"},
            {"name": "topic_score", "type": "float", "description": "Topic relevance score (0-1)"},
            {"name": "key_phrases", "type": "str", "description": "Comma-separated list of key phrases identified"},
            {"name": "topics", "type": "str", "description": "Comma-separated list of topics discussed"},
            {"name": "emotions", "type": "str", "description": "Emotional analysis metrics"},
            {"name": "confidence_score", "type": "str", "description": "AI confidence score for analytics"},
        ],
        "filterable_columns": ["date"],
        "api_endpoint": "/v2/calls/extensive",
        "supports_pagination": True,
        "notes": "Defaults to last 7 days if no date filters are provided.",
    },
    "transcripts": {
        "name": "transcripts",
        "type": "api_table",
        "description": "Call transcripts with speaker identification, timestamps, and confidence scores",
        "columns": [
            {"name": "call_id", "type": "str", "description": "Unique identifier for the call"},
            {"name": "speaker", "type": "str", "description": "Speaker identifier"},
            {"name": "timestamp", "type": "int", "description": "Timestamp in milliseconds from call start"},
            {"name": "text", "type": "str", "description": "Transcript text for this segment"},
            {"name": "confidence", "type": "float", "description": "Transcription confidence score"},
            {"name": "segment_id", "type": "str", "description": "Unique segment identifier"},
        ],
        "filterable_columns": ["call_id", "text"],
        "api_endpoint": "/v2/calls/transcript",
        "supports_pagination": True,
        "notes": "Fetch transcripts for specific call IDs; always filter by call_id and/or narrow date range.",
    },
}


GONG_PRIMARY_KEYS = {
    "calls": {"column_name": "call_id", "constraint_name": "pk_calls_call_id"},
    "users": {"column_name": "user_id", "constraint_name": "pk_users_user_id"},
    "analytics": {"column_name": "call_id", "constraint_name": "pk_analytics_call_id"},
    "transcripts": {"column_name": "segment_id", "constraint_name": "pk_transcripts_segment_id"},
}

GONG_FOREIGN_KEYS = {
    "analytics": [
        {
            "column_name": "call_id",
            "foreign_table_name": "calls",
            "foreign_column_name": "call_id",
            "constraint_name": "fk_analytics_call_id",
        }
    ],
    "transcripts": [
        {
            "column_name": "call_id",
            "foreign_table_name": "calls",
            "foreign_column_name": "call_id",
            "constraint_name": "fk_transcripts_call_id",
        }
    ],
}



import sys
import time
import threading
from dataclasses import dataclass, field
from collections import defaultdict

from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities import log

logger = log.getLogger(__name__)


@dataclass(kw_only=True, slots=True)
class HandlersCacheRecord:
    """Record for a handler in the cache

    Args:
        handler (DatabaseHandler): handler instance
        expired_at (float): time when the handler will be expired
    """

    handler: DatabaseHandler
    expired_at: float
    connect_attempt_done: threading.Event = field(default_factory=threading.Event)

    @property
    def expired(self) -> bool:
        """check if the handler is expired

        Returns:
            bool: True if the handler is expired, False otherwise
        """
        return self.expired_at < time.time()

    @property
    def has_references(self) -> bool:
        """check if the handler has references

        Returns:
            bool: True if the handler has references, False otherwise
        """
        return sys.getrefcount(self.handler) > 2

    def connect(self) -> None:
        """connect to the handler"""
        try:
            if not self.handler.is_connected:
                self.handler.connect()
        except Exception:
            logger.warning(f"Error connecting to handler: {self.handler.name}", exc_info=True)
        finally:
            self.connect_attempt_done.set()


class HandlersCache:
    """Cache for data handlers that keep connections opened during ttl time from handler last use"""

    def __init__(self, ttl: int = 60, clean_timeout: float = 3):
        """init cache

        Args:
            ttl (int): time to live (in seconds) for record in cache
            clean_timeout (float): interval between cleanups of expired handlers
        """
        self.ttl: int = ttl
        self._clean_timeout: int = clean_timeout
        self.handlers: dict[str, list[HandlersCacheRecord]] = defaultdict(list)
        self._lock = threading.RLock()
        self._stop_event = threading.Event()
        self.cleaner_thread = None

    def __del__(self):
        self._stop_clean()

    def _start_clean(self) -> None:
        """start worker that close connections after ttl expired"""
        if isinstance(self.cleaner_thread, threading.Thread) and self.cleaner_thread.is_alive():
            return
        with self._lock:
            self._stop_event.clear()
            self.cleaner_thread = threading.Thread(target=self._clean, name="HandlersCache.clean")
            self.cleaner_thread.daemon = True
            self.cleaner_thread.start()

    def _stop_clean(self) -> None:
        """stop clean worker"""
        self._stop_event.set()

    def set(self, handler: DatabaseHandler):
        """add (or replace) handler in cache

        NOTE: If the handler is not thread-safe, then use a lock when making connection. Otherwise, make connection in
        the same thread without using a lock to speed up parallel queries. (They don't need to wait for a connection in
        another thread.)

        Args:
            handler (DatabaseHandler)
        """
        thread_safe = getattr(handler, "thread_safe", True)
        with self._lock:
            try:
                # If the handler is defined to be thread safe, set 0 as the last element of the key, otherwise set the thrad ID.
                key = (
                    handler.name,
                    ctx.company_id,
                    0 if thread_safe else threading.get_native_id(),
                )
                record = HandlersCacheRecord(handler=handler, expired_at=time.time() + self.ttl)
                self.handlers[key].append(record)
            except Exception:
                logger.warning("Error setting data handler cache record:", exc_info=True)
                return
        self._start_clean()
        record.connect()

    def _get_cache_records(self, name: str) -> tuple[list[HandlersCacheRecord] | None, str]:
        """get cache records by name

        Args:
            name (str): handler name

        Returns:
            tuple[list[HandlersCacheRecord] | None, str]: cache records and key of the handler in cache
        """
        # If the handler is not thread safe, the thread ID will be assigned to the last element of the key.
        key = (name, ctx.company_id, 0)
        if key not in self.handlers:
            key = (name, ctx.company_id, threading.get_native_id())
        return self.handlers.get(key, []), key

    def get(self, name: str) -> DatabaseHandler | None:
        """get handler from cache by name

        Args:
            name (str): handler name

        Returns:
            DatabaseHandler
        """
        with self._lock:
            records, _ = self._get_cache_records(name)
            for record in records:
                if record.expired is False and record.has_references is False:
                    record.expired_at = time.time() + self.ttl
                    if record.connect_attempt_done.wait(timeout=10) is False:
                        logger.warning(f"Handler's connection attempt has not finished in 10s: {record.handler.name}")
                    return record.handler
            return None

    def delete(self, name: str) -> None:
        """delete handler from cache

        Args:
            name (str): handler name
        """
        with self._lock:
            records, key = self._get_cache_records(name)
            if len(records) > 0:
                del self.handlers[key]
            for record in records:
                try:
                    record.handler.disconnect()
                except Exception:
                    logger.debug("Error disconnecting data handler:", exc_info=True)

            if len(self.handlers) == 0:
                self._stop_clean()

    def _clean(self) -> None:
        """worker that delete from cache handlers that was not in use for ttl"""
        while self._stop_event.wait(timeout=self._clean_timeout) is False:
            with self._lock:
                for key in list(self.handlers.keys()):
                    active_handlers_list = []
                    for record in self.handlers[key]:
                        if record.expired and record.has_references is False:
                            try:
                                record.handler.disconnect()
                            except Exception:
                                logger.debug("Error disconnecting data handler:", exc_info=True)
                        else:
                            active_handlers_list.append(record)
                    if len(active_handlers_list) > 0:
                        self.handlers[key] = active_handlers_list
                    else:
                        del self.handlers[key]

                if len(self.handlers) == 0:
                    self._stop_event.set()



from datetime import datetime, date, timedelta
from decimal import Decimal
import pandas as pd
import numpy as np
import orjson
from flask.json.provider import DefaultJSONProvider


class CustomJSONEncoder:
    def default(self, obj):
        if isinstance(obj, timedelta):
            return str(obj)
        if isinstance(obj, datetime):
            return obj.strftime("%Y-%m-%d %H:%M:%S.%f")
        if isinstance(obj, date):
            return obj.strftime("%Y-%m-%d")
        if isinstance(obj, Decimal):
            return float(obj)
        if isinstance(obj, np.bool_):
            return bool(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if pd.isnull(obj):
            return None

        return str(obj)


class ORJSONProvider(DefaultJSONProvider):
    """
    Use orjson to serialize data instead of flask json provider.
    """

    def dumps(self, obj, **kwargs):
        return orjson.dumps(
            obj,
            option=(
                orjson.OPT_SERIALIZE_NUMPY
                | orjson.OPT_NON_STR_KEYS
                # keep this for using CustomJSON encoder
                | orjson.OPT_PASSTHROUGH_DATETIME
            ),
            default=CustomJSONEncoder().default,
        ).decode("utf-8")

    def loads(self, s, **kwargs):
        return orjson.loads(s)



from typing import Any

from mindsdb.integrations.handlers.bigcommerce_handler.bigcommerce_api_client import BigCommerceAPIClient
from mindsdb.integrations.handlers.bigcommerce_handler.bigcommerce_tables import (
    BigCommerceOrdersTable,
    BigCommerceProductsTable,
    BigCommerceCustomersTable,
    BigCommerceCategoriesTable,
    BigCommercePickupsTable,
    BigCommercePromotionsTable,
    BigCommerceWishlistsTable,
    BigCommerceSegmentsTable,
    BigCommerceBrandsTable,
)
from mindsdb.integrations.libs.api_handler import MetaAPIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
)
from mindsdb.utilities import log


logger = log.getLogger(__name__)


class BigCommerceHandler(MetaAPIHandler):
    """This handler handles the connection and execution of SQL statements on BigCommerce."""

    name = "bigcommerce"

    def __init__(self, name: str, connection_data: dict, **kwargs: Any) -> None:
        """
        Initializes the handler.

        Args:
            name (str): The name of the handler instance.
            connection_data (dict): The connection data required to connect to the BigCommerce API.
            kwargs: Arbitrary keyword arguments.
        """
        super().__init__(name)
        self.connection_data = connection_data
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False
        self.thread_safe = True

        self._register_table("orders", BigCommerceOrdersTable(self))
        self._register_table("products", BigCommerceProductsTable(self))
        self._register_table("customers", BigCommerceCustomersTable(self))
        self._register_table("categories", BigCommerceCategoriesTable(self))
        self._register_table("pickups", BigCommercePickupsTable(self))
        self._register_table("promotions", BigCommercePromotionsTable(self))
        self._register_table("wishlists", BigCommerceWishlistsTable(self))
        self._register_table("segments", BigCommerceSegmentsTable(self))
        self._register_table("brands", BigCommerceBrandsTable(self))

    def connect(self) -> BigCommerceAPIClient:
        """
        Establishes a connection to the BigCommerce API.

        Raises:
            ValueError: If the required connection parameters are not provided.

        Returns:
            BigCommerceAPIClient: A connection object to the BigCommerce API.
        """
        if self.is_connected is True:
            return self.connection

        if not all(
            key in self.connection_data and self.connection_data.get(key) for key in ["api_base", "access_token"]
        ):
            raise ValueError("Required parameters (api_base, access_token) must be provided and should not be empty.")

        self.connection = BigCommerceAPIClient(
            url=self.connection_data.get("api_base"),
            access_token=self.connection_data.get("access_token"),
        )

        self.is_connected = True
        return self.connection

    def check_connection(self) -> StatusResponse:
        """
        Checks the status of the connection to the BigCommerce API.

        Returns:
            StatusResponse: An object containing the success status and an error message if an error occurs.
        """
        response = StatusResponse(False)

        try:
            connection = self.connect()
            connection.get_products(limit=1)
            response.success = True
        except Exception as e:
            logger.error(f"Error connecting to BigCommerce API: {e}!")
            response.error_message = e

        self.is_connected = response.success

        return response



from http import HTTPStatus

import json

from flask import request
from flask_restx import Resource
import pandas as pd

from mindsdb.api.http.namespaces.configs.projects import ns_conf
from mindsdb.api.http.utils import http_error
from mindsdb.api.executor.controllers.session_controller import SessionController
from mindsdb.metrics.metrics import api_endpoint_metrics
from mindsdb.interfaces.model.functions import PredictorRecordNotFound
from mindsdb.interfaces.storage.db import Predictor
from mindsdb_sql_parser import parse_sql
from mindsdb_sql_parser.ast.mindsdb import CreatePredictor
from mindsdb.utilities.exception import EntityNotExistsError
from mindsdb.utilities import log

logger = log.getLogger(__name__)


@ns_conf.route("/<project_name>/models")
class ModelsList(Resource):
    @ns_conf.doc("list_models")
    @api_endpoint_metrics("GET", "/models")
    def get(self, project_name):
        """List all models"""
        session = SessionController()

        try:
            session.database_controller.get_project(project_name)
        except EntityNotExistsError:
            return http_error(HTTPStatus.NOT_FOUND, "Project not found", f"Project name {project_name} does not exist")

        return session.model_controller.get_models(with_versions=True, project_name=project_name)

    @ns_conf.doc("train_model")
    @api_endpoint_metrics("POST", "/models")
    def post(self, project_name):
        """Creates a new model and trains it"""
        session = SessionController()

        if "query" not in request.json:
            return http_error(HTTPStatus.BAD_REQUEST, "Query required", 'Missing "query" SQL statement')
        query = request.json["query"]

        project_datanode = session.datahub.get(project_name)
        if project_datanode is None:
            return http_error(HTTPStatus.NOT_FOUND, "Project not found", f"Project name {project_name} does not exist")

        try:
            create_statement = parse_sql(query)
        except Exception:
            return http_error(
                HTTPStatus.BAD_REQUEST, "Invalid query string", f"SQL CREATE statement is invalid: {query}"
            )

        if type(create_statement) is not CreatePredictor:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                "Invalid CREATE SQL statement",
                f"SQL statement is not a CREATE model statement: {query}",
            )

        model_name = create_statement.name.parts[1].lower()
        try:
            session.model_controller.get_model(model_name, project_name=project_name)
            return http_error(
                HTTPStatus.CONFLICT, "Model already exists", f"Model with name {model_name} already exists"
            )
        except PredictorRecordNotFound:
            pass

        ml_integration = "lightwood"
        if create_statement.using is not None:
            # Convert using to lowercase
            create_statement.using = {k.lower(): v for k, v in create_statement.using.items()}
            ml_integration = create_statement.using.pop("engine", ml_integration)

        try:
            ml_handler = session.integration_controller.get_ml_handler(ml_integration)
        except Exception:
            return http_error(
                HTTPStatus.NOT_FOUND, "ML handler not found", f"Cannot find ML handler with name {ml_integration}"
            )

        try:
            model_df = session.model_controller.create_model(create_statement, ml_handler)
            # Consistent format with GET /projects/<project_name>/models/<model_name>
            return {
                "name": model_df.at[0, "NAME"],
                "accuracy": None,
                "active": model_df.at[0, "ACTIVE"],
                "version": model_df.at[0, "VERSION"],
                "status": model_df.at[0, "STATUS"],
                "predict": model_df.at[0, "PREDICT"],
                "mindsdb_version": model_df.at[0, "MINDSDB_VERSION"],
                "error": model_df.at[0, "ERROR"],
                "fetch_data_query": model_df.at[0, "SELECT_DATA_QUERY"],
                "problem_definition": model_df.at[0, "TRAINING_OPTIONS"],
            }, HTTPStatus.CREATED
        except Exception as e:
            logger.exception("Something went wrong while creating and training model")
            return http_error(
                HTTPStatus.INTERNAL_SERVER_ERROR,
                "Unable to train model",
                f"Something went wrong while creating and training model {model_name}: {e}",
            )


@ns_conf.route("/<project_name>/models/<model_name>")
@ns_conf.param("project_name", "Name of the project")
@ns_conf.param("model_name", "Name of the model")
class ModelResource(Resource):
    @ns_conf.doc("get_model")
    @api_endpoint_metrics("GET", "/models/model")
    def get(self, project_name, model_name):
        """Get a model by name and version"""
        session = SessionController()

        project_datanode = session.datahub.get(project_name)
        if project_datanode is None:
            return http_error(HTTPStatus.NOT_FOUND, "Project not found", f"Project name {project_name} does not exist")

        name_no_version, version = Predictor.get_name_and_version(model_name)
        try:
            return session.model_controller.get_model(name_no_version, version=version, project_name=project_name)
        except PredictorRecordNotFound:
            return http_error(HTTPStatus.NOT_FOUND, "Model not found", f"Model with name {model_name} not found")

    @ns_conf.doc("update_model")
    @api_endpoint_metrics("PUT", "/models/model")
    def put(self, project_name, model_name):
        """Update model"""

        session = SessionController()

        project_datanode = session.datahub.get(project_name)
        if project_datanode is None:
            return http_error(HTTPStatus.NOT_FOUND, "Project not found", f"Project name {project_name} does not exist")

        if "problem_definition" not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST, "problem_definition required", 'Missing "problem_definition" field'
            )

        problem_definition = request.json["problem_definition"]

        model_name, version = Predictor.get_name_and_version(model_name)

        session.model_controller.update_model(
            session,
            project_name,
            model_name,
            version=version,
            problem_definition=problem_definition,
        )
        return session.model_controller.get_model(model_name, version=version, project_name=project_name)

    @ns_conf.doc("delete_model")
    @api_endpoint_metrics("DELETE", "/models/model")
    def delete(self, project_name, model_name):
        """Deletes a model by name"""

        session = SessionController()

        project_datanode = session.datahub.get(project_name)
        if project_datanode is None:
            return http_error(HTTPStatus.NOT_FOUND, "Project not found", f"Project name {project_name} does not exist")

        name_no_version, version = Predictor.get_name_and_version(model_name)
        try:
            session.model_controller.get_model(name_no_version, version=version, project_name=project_name)
        except PredictorRecordNotFound:
            return http_error(HTTPStatus.NOT_FOUND, "Model not found", f"Model with name {model_name} not found")

        try:
            session.model_controller.delete_model(name_no_version, project_name, version=version)
        except Exception as e:
            logger.exception(f"Something went wrong while deleting model '{model_name}'")
            return http_error(
                HTTPStatus.INTERNAL_SERVER_ERROR,
                "Error deleting model",
                f"Something went wrong while deleting {model_name}: {e}",
            )

        return "", HTTPStatus.NO_CONTENT


@ns_conf.route("/<project_name>/models/<model_name>/predict")
@ns_conf.param("project_name", "Name of the project")
@ns_conf.param("model_name", "Name of the model")
class ModelPredict(Resource):
    @ns_conf.doc("post_model_predict")
    @api_endpoint_metrics("POST", "/models/model/predict")
    def post(self, project_name, model_name):
        """Call prediction"""

        name_no_version, version = Predictor.get_name_and_version(model_name)

        session = SessionController()
        project_datanode = session.datahub.get(project_name)
        if project_datanode is None:
            return http_error(
                HTTPStatus.NOT_FOUND, "Project not found", f"Project with name {project_name} does not exist"
            )

        try:
            session.model_controller.get_model(name_no_version, version=version, project_name=project_name)
        except PredictorRecordNotFound:
            return http_error(HTTPStatus.NOT_FOUND, "Model not found", f"Model with name {model_name} not found")

        data = request.json["data"]
        if isinstance(data, str):
            # Support object or serialized object.
            data = json.loads(data)
        params = request.json.get("params")

        predictions = project_datanode.predict(
            model_name=name_no_version,
            df=pd.DataFrame(data),
            version=version,
            params=params,
        )

        return predictions.to_dict("records")


@ns_conf.route("/<project_name>/models/<model_name>/describe")
@ns_conf.param("project_name", "Name of the project")
@ns_conf.param("model_name", "Name of the model")
class ModelDescribe(Resource):
    @ns_conf.doc("describe_model")
    @api_endpoint_metrics("GET", "/models/model/describe")
    def get(self, project_name, model_name):
        """Describes a model"""
        session = SessionController()

        project_datanode = session.datahub.get(project_name)
        if project_datanode is None:
            return http_error(HTTPStatus.NOT_FOUND, "Project not found", f"Project name {project_name} does not exist")

        name_no_version, version = Predictor.get_name_and_version(model_name)

        try:
            session.model_controller.get_model(name_no_version, version=version, project_name=project_name)
        except PredictorRecordNotFound:
            return http_error(HTTPStatus.NOT_FOUND, "Model not found", f"Model with name {model_name} not found")

        attribute = None
        if "attribute" in request.json:
            attribute = request.json["attribute"]

        try:
            description_df = session.model_controller.describe_model(
                session, project_name, name_no_version, attribute, version=version
            )
            return description_df.to_dict("records")
        except Exception:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                "ML handler unsupported",
                f"ML handler for {model_name} does not support model description",
            )



from typing import Optional, List

from sqlalchemy import null, func

import mindsdb.interfaces.storage.db as db
from mindsdb.utilities.context import context as ctx
import mindsdb.utilities.profiler as profiler


class PredictorRecordNotFound(Exception):
    def __init__(self, **kwargs):
        name = kwargs.get("name") or "-"
        predictor_id = kwargs.get("id") or "-"
        super().__init__(f"Predictor not found: name='{name}' id='{predictor_id}'")


class MultiplePredictorRecordsFound(Exception):
    def __init__(self, **kwargs):
        name = kwargs.get("name") or "-"
        predictor_id = kwargs.get("id") or "-"
        super().__init__(f"Found multiple predictor with: name='{name}' id='{predictor_id}'")


@profiler.profile()
def get_integration_record(name: str) -> db.Integration:
    company_id = ctx.company_id
    if company_id is None:
        company_id = null()

    record = db.session.query(db.Integration).filter_by(company_id=company_id, name=name).first()
    return record


@profiler.profile()
def get_project_record(name: str) -> db.Project:
    company_id = ctx.company_id if ctx.company_id is not None else "0"

    project_record = (
        db.session.query(db.Project)
        .filter(
            (func.lower(db.Project.name) == name)
            & (db.Project.company_id == company_id)
            & (db.Project.deleted_at == null())
        )
        .first()
    )
    return project_record


@profiler.profile()
def get_project_records() -> List[db.Project]:
    company_id = ctx.company_id if ctx.company_id is not None else "0"

    return (
        db.session.query(db.Project)
        .filter((db.Project.company_id == company_id) & (db.Project.deleted_at == null()))
        .all()
    )


@profiler.profile()
def get_predictor_integration(record: db.Predictor) -> db.Integration:
    integration_record = db.session.query(db.Integration).filter_by(id=record.integration_id).first()
    return integration_record


@profiler.profile()
def get_predictor_project(record: db.Predictor) -> db.Project:
    project_record = db.session.query(db.Project).filter_by(id=record.project_id).first()
    return project_record


@profiler.profile()
def get_model_records(
    integration_id=None,
    active=True,
    deleted_at=null(),
    project_name: Optional[str] = None,
    ml_handler_name: Optional[str] = None,
    **kwargs,
):
    kwargs["company_id"] = ctx.company_id
    if kwargs["company_id"] is None:
        kwargs["company_id"] = null()

    if deleted_at is not None:
        kwargs["deleted_at"] = deleted_at
    if active is not None:
        kwargs["active"] = active

    if project_name is not None:
        project_record = get_project_record(name=project_name)
        if project_record is None:
            return []
        kwargs["project_id"] = project_record.id

    if ml_handler_name is not None:
        ml_handler_record = get_integration_record(name=ml_handler_name)
        if ml_handler_record is None:
            # raise Exception(f'unknown ml handler: {ml_handler_name}')
            return []
        kwargs["integration_id"] = ml_handler_record.id

    if integration_id is not None:
        kwargs["integration_id"] = integration_id

    return db.session.query(db.Predictor).filter_by(**kwargs).all()


@profiler.profile()
def get_model_record(
    except_absent=False,
    ml_handler_name: Optional[str] = None,
    project_name: Optional[str] = None,
    active: bool = True,
    deleted_at=null(),
    version: Optional[int] = None,
    **kwargs,
):
    kwargs["company_id"] = ctx.company_id
    if kwargs["company_id"] is None:
        kwargs["company_id"] = null()

    kwargs["deleted_at"] = deleted_at
    if active is not None:
        kwargs["active"] = active
    if version is not None:
        kwargs["version"] = version

    if project_name is not None:
        project_record = get_project_record(name=project_name)
        if project_record is None:
            return None
        kwargs["project_id"] = project_record.id

    if ml_handler_name is not None:
        ml_handler_record = get_integration_record(name=ml_handler_name)
        if ml_handler_record is None:
            # raise Exception(f'unknown ml handler: {ml_handler_name}')
            return []
        kwargs["integration_id"] = ml_handler_record.id

    records = db.session.query(db.Predictor).filter_by(**kwargs).all()
    if len(records) > 1:
        raise MultiplePredictorRecordsFound(**kwargs)
    if len(records) == 0:
        if except_absent is True:
            raise PredictorRecordNotFound(**kwargs)
        else:
            return None
    return records[0]



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    account={
        "type": ARG_TYPE.STR,
        "description": "The Snowflake account identifier.",
        "required": True,
        "label": "Account",
    },
    user={
        "type": ARG_TYPE.STR,
        "description": "The user name used to authenticate with the Snowflake account.",
        "required": True,
        "label": "User",
    },
    password={
        "type": ARG_TYPE.PWD,
        "description": "The password to authenticate the user with the Snowflake account. Required for password authentication.",
        "required": False,
        "label": "Password",
        "secret": True,
    },
    private_key_path={
        "type": ARG_TYPE.PATH,
        "description": "Path to the private key file for key pair authentication. Required for key pair authentication.",
        "required": False,
        "label": "Private Key Path",
    },
    private_key_passphrase={
        "type": ARG_TYPE.PWD,
        "description": "Optional passphrase for the encrypted private key.",
        "required": False,
        "label": "Private Key Passphrase",
        "secret": True,
    },
    database={
        "type": ARG_TYPE.STR,
        "description": "The database to use when connecting to the Snowflake account.",
        "required": True,
        "label": "Database",
    },
    schema={
        "type": ARG_TYPE.STR,
        "description": "The schema to use when connecting to the Snowflake account.",
        "required": False,
        "label": "Schema",
    },
    warehouse={
        "type": ARG_TYPE.STR,
        "description": "The warehouse to use when executing queries on the Snowflake account.",
        "required": False,
        "label": "Warehouse",
    },
    role={
        "type": ARG_TYPE.STR,
        "description": "The role to use when executing queries on the Snowflake account.",
        "required": False,
        "label": "Role",
    },
    auth_type={
        "type": ARG_TYPE.STR,
        "description": 'Required authentication type. Options: "password" or "key_pair".',
        "required": True,
        "label": "Auth Type",
    },
)

connection_args_example = OrderedDict(
    account="abcxyz-1234567", user="user", password="password", database="test", auth_type="password"
)



from http import HTTPStatus
from urllib.parse import urljoin

import requests


DEFAULT_LIMIT = 999999999


class BigCommerceAPIClient:
    def __init__(self, url: str, access_token: str):
        # we have to use both endpoints: v2/ and v3/, so delete it from base url
        self.base_url = url.rstrip("/")
        self.base_url = self.base_url.rstrip("v2")
        self.base_url = self.base_url.rstrip("v3")
        if not self.base_url.endswith("/"):
            self.base_url += "/"

        self.access_token = access_token
        self.session = requests.Session()
        self.session.headers.update(
            {
                "X-Auth-Token": self.access_token,
                "Content-Type": "application/json",
                "Accept": "application/json",
            }
        )

    def get_products(
        self,
        filter: dict = None,
        sort_condition: dict = None,
        limit: int = None,
    ):
        # doc: https://developer.bigcommerce.com/docs/rest-catalog/products#get-all-products
        params = {
            "limit": limit or DEFAULT_LIMIT,
        }
        if filter is not None:
            params.update(filter)
        if sort_condition is not None:
            params.update(sort_condition)
        return self._make_request_v3("GET", "catalog/products", params=params)

    def get_customers(
        self,
        filter: dict = None,
        sort_condition: dict = None,
        limit: int = None,
    ):
        # doc: https://developer.bigcommerce.com/docs/rest-management/customers#get-all-customers
        params = {
            "limit": limit or DEFAULT_LIMIT,
        }
        if filter:
            params.update(filter)
        if sort_condition:
            params["sort"] = sort_condition
        return self._make_request_v3("GET", "customers", params=params)

    def get_orders(
        self,
        filter: dict = None,
        sort_condition: str = None,
        limit: int = None,
    ):
        # doc: https://developer.bigcommerce.com/docs/rest-management/orders#get-all-orders
        params = {"limit": limit or DEFAULT_LIMIT}
        if filter:
            params.update(filter)
        if sort_condition:
            params["sort"] = sort_condition

        return self._make_request_v2("GET", "orders", params=params)

    def get_orders_count(self) -> int:
        response = self._make_request_v2("GET", "orders/count")
        return response["count"]

    def get_customers_count(self) -> int:
        response = self._make_request("GET", urljoin(self.base_url, "v3/customers"), params={"limit": 1})
        return response["meta"]["pagination"]["total"]

    def get_products_count(self) -> int:
        response = self._make_request("GET", urljoin(self.base_url, "v3/products"), params={"limit": 1})
        return response["meta"]["pagination"]["total"]

    def get_categories(
        self,
        filter: dict = None,
        limit: int = None,
    ):
        # doc: https://developer.bigcommerce.com/docs/rest-catalog/category-trees/categories#get-all-categories
        params = {
            "limit": limit or DEFAULT_LIMIT,
        }
        if filter is not None:
            params.update(filter)
        return self._make_request_v3("GET", "catalog/trees/categories", params=params)

    def get_categories_count(self) -> int:
        response = self._make_request("GET", urljoin(self.base_url, "v3/catalog/trees/categories"), params={"limit": 1})
        return response.get("meta", {}).get("pagination", {}).get("total", 0)

    def get_pickups(
        self,
        filter: dict = None,
        limit: int = None,
    ):
        # doc: https://developer.bigcommerce.com/docs/rest-management/pickup#get-pickups
        params = {
            "limit": limit or DEFAULT_LIMIT,
        }
        if filter is not None:
            params.update(filter)
        return self._make_request_v3("GET", "orders/pickups", params=params)

    def get_pickups_count(self) -> int:
        response = self._make_request("GET", urljoin(self.base_url, "v3/pickups"), params={"limit": 1})
        return response.get("meta", {}).get("pagination", {}).get("total", 0)

    def get_promotions(
        self,
        filter: dict = None,
        sort_condition: dict = None,
        limit: int = None,
    ):
        # doc: https://developer.bigcommerce.com/docs/rest-management/promotions/promotions-bulk#get-all-promotions
        params = {
            "limit": limit or DEFAULT_LIMIT,
        }
        if filter is not None:
            params.update(filter)
        if sort_condition is not None:
            params.update(sort_condition)
        return self._make_request_v3("GET", "promotions", params=params)

    def get_promotions_count(self) -> int:
        response = self._make_request("GET", urljoin(self.base_url, "v3/promotions"), params={"limit": 1})
        return response.get("meta", {}).get("pagination", {}).get("total", 0)

    def get_wishlists(
        self,
        filter: dict = None,
        limit: int = None,
    ):
        # doc: https://developer.bigcommerce.com/docs/rest-management/wishlists#get-all-wishlists
        params = {
            "limit": limit or DEFAULT_LIMIT,
        }
        if filter is not None:
            params.update(filter)
        return self._make_request_v3("GET", "wishlists", params=params)

    def get_wishlists_count(self) -> int:
        response = self._make_request("GET", urljoin(self.base_url, "v3/wishlists"), params={"limit": 1})
        return response.get("meta", {}).get("pagination", {}).get("total", 0)

    def get_segments(
        self,
        filter: dict = None,
        limit: int = None,
    ):
        # doc: https://developer.bigcommerce.com/docs/rest-management/customer-segmentation/segments#get-all-segments
        params = {
            "limit": limit or DEFAULT_LIMIT,
        }
        if filter is not None:
            params.update(filter)
        return self._make_request_v3("GET", "segments", params=params)

    def get_segments_count(self) -> int:
        response = self._make_request("GET", urljoin(self.base_url, "v3/segments"), params={"limit": 1})
        return response.get("meta", {}).get("pagination", {}).get("total", 0)

    def get_brands(
        self,
        filter: dict = None,
        sort_condition: dict = None,
        limit: int = None,
    ):
        # doc: https://developer.bigcommerce.com/docs/rest-catalog/brands#get-all-brands
        params = {
            "limit": limit or DEFAULT_LIMIT,
        }
        if filter is not None:
            params.update(filter)
        if sort_condition is not None:
            params.update(sort_condition)
        return self._make_request_v3("GET", "catalog/brands", params=params)

    def get_brands_count(self) -> int:
        response = self._make_request("GET", urljoin(self.base_url, "v3/catalog/brands"), params={"limit": 1})
        return response.get("meta", {}).get("pagination", {}).get("total", 0)

    def _make_request_v2(self, method: str, url: str, *args, **kwargs) -> list[dict]:
        # NOTE: v2 limit max is 250
        url = urljoin(urljoin(self.base_url, "v2/"), url)
        api_limit = 250
        params = kwargs.pop("params", {})
        request_limit = params.get("limit", DEFAULT_LIMIT)
        params["limit"] = min(api_limit, request_limit)
        current_page = 1
        response_len = 1
        data = []
        while response_len > 0 and len(data) < request_limit:
            params["page"] = current_page
            response = self._make_request(method, url, params=params, *args, **kwargs)
            if isinstance(response, dict):
                # for "get_count" requests
                return response
            current_page += 1
            response_len = len(response)
            data += response
        return data[:request_limit]

    def _make_request_v3(self, method: str, url: str, *args, **kwargs) -> list[dict]:
        url = urljoin(urljoin(self.base_url, "v3/"), url)
        data = []
        params = kwargs.pop("params", {})
        current_page = 1
        total_pages = 1
        while current_page <= total_pages:
            params["page"] = current_page
            response = self._make_request(method, url, params=params, *args, **kwargs)
            current_page = response["meta"]["pagination"]["current_page"] + 1
            total_pages = response["meta"]["pagination"]["total_pages"]
            data += response["data"]
        return data

    def _make_request(self, method: str, url: str, params: dict = None, data: dict = None) -> dict:
        response = self.session.request(method, url, params=params, json=data)

        if response.status_code == HTTPStatus.NO_CONTENT:
            return []
        if response.status_code != HTTPStatus.OK:
            raise Exception(f"Request failed with status code {response.status_code}: {response.text}")

        return response.json()



from contextvars import ContextVar
from typing import Any
from copy import deepcopy


class Context:
    """Thread independent storage"""

    __slots__ = ("_storage",)

    def __init__(self, storage) -> None:
        object.__setattr__(self, "_storage", storage)
        self.set_default()

    def set_default(self) -> None:
        self._storage.set(
            {
                "user_id": None,
                "company_id": None,
                "session_id": "",
                "task_id": None,
                "user_class": 0,
                "profiling": {"level": 0, "enabled": False, "pointer": None, "tree": None},
                "email_confirmed": 0,
                "used_handlers": set(),
            }
        )

    def __getattr__(self, name: str) -> Any:
        storage = self._storage.get({})
        if name not in storage:
            raise AttributeError(name)
        return storage[name]

    def __setattr__(self, name: str, value: Any) -> None:
        storage = deepcopy(self._storage.get({}))
        storage[name] = value
        self._storage.set(storage)

    def __delattr__(self, name: str) -> None:
        storage = deepcopy(self._storage.get({}))
        if name not in storage:
            raise AttributeError(name)
        del storage["name"]
        self._storage.set(storage)

    def dump(self) -> dict:
        storage = deepcopy(self._storage.get({}))
        return storage

    def load(self, storage: dict) -> None:
        self._storage.set(storage)

    def get_metadata(self, **kwargs) -> dict:
        return {
            "user_id": self.user_id or "",
            "company_id": self.company_id or "",
            "session_id": self.session_id,
            "user_class": self.user_class,
            **kwargs,
        }


_context_var = ContextVar("mindsdb.context")
context = Context(_context_var)



from abc import ABC, abstractmethod
from typing import Dict, Any
from pathlib import Path


class SnowflakeAuthType(ABC):
    @abstractmethod
    def get_config(self, **kwargs) -> Dict[str, Any]:
        pass


class PasswordAuthType(SnowflakeAuthType):
    def get_config(self, **kwargs) -> Dict[str, Any]:
        required_keys = ["account", "user", "database"]
        if not all(kwargs.get(key) for key in required_keys):
            raise ValueError("Required parameters (account, user, database) must be provided.")

        if not kwargs.get("password"):
            raise ValueError("Password must be provided when auth_type is 'password'.")
        return {
            "account": kwargs.get("account"),
            "user": kwargs.get("user"),
            "password": kwargs.get("password"),
            "database": kwargs.get("database"),
            "schema": kwargs.get("schema"),
            "role": kwargs.get("role"),
            "warehouse": kwargs.get("warehouse"),
            "auth_type": "password",
        }


class KeyPairAuthType(SnowflakeAuthType):
    def get_config(self, **kwargs) -> Dict[str, Any]:
        if not kwargs.get("private_key_path"):
            raise ValueError("private_key_path must be provided when auth_type is 'key_pair'.")

        if not all(kwargs.get(key) for key in ["account", "user", "database"]):
            raise ValueError("Required parameters (account, user, database) must be provided.")

        private_key_path = kwargs.get("private_key_path")
        if not Path(private_key_path).exists():
            raise ValueError(f"Private key file not found: {private_key_path}")

        config = {
            "account": kwargs.get("account"),
            "user": kwargs.get("user"),
            "private_key_file": private_key_path,
            "database": kwargs.get("database"),
            "schema": kwargs.get("schema"),
            "role": kwargs.get("role"),
            "warehouse": kwargs.get("warehouse"),
            "authenticator": "SNOWFLAKE_JWT",
            "auth_type": "key_pair",
        }
        if kwargs.get("private_key_passphrase"):
            config["private_key_file_pwd"] = kwargs.get("private_key_passphrase")
        return config



import copy
import datetime as dt
from copy import deepcopy
from multiprocessing.pool import ThreadPool

import pandas as pd
from dateutil.parser import parse as parse_datetime

from sqlalchemy import func
import numpy as np

import mindsdb.interfaces.storage.db as db
from mindsdb.utilities.config import Config
from mindsdb.interfaces.model.functions import get_model_record, get_model_records, get_project_record
from mindsdb.interfaces.storage.json import get_json_storage
from mindsdb.interfaces.storage.model_fs import ModelStorage
from mindsdb.utilities.config import config
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities.functions import resolve_model_identifier
import mindsdb.utilities.profiler as profiler
from mindsdb.utilities.exception import EntityExistsError, EntityNotExistsError
from mindsdb.utilities import log

logger = log.getLogger(__name__)


default_project = config.get("default_project")


def delete_model_storage(model_id, ctx_dump):
    try:
        ctx.load(ctx_dump)
        modelStorage = ModelStorage(model_id)
        modelStorage.delete()
    except Exception:
        logger.exception(f"Something went wrong during deleting storage of model {model_id}:")


class ModelController:
    config: Config

    def __init__(self) -> None:
        self.config = Config()

    def get_model_data(self, name: str = None, predictor_record=None, ml_handler_name="lightwood") -> dict:
        if predictor_record is None:
            predictor_record = get_model_record(except_absent=True, name=name, ml_handler_name=ml_handler_name)

        data = deepcopy(predictor_record.data)
        data["dtype_dict"] = predictor_record.dtype_dict
        data["created_at"] = str(parse_datetime(str(predictor_record.created_at).split(".")[0]))
        data["updated_at"] = str(parse_datetime(str(predictor_record.updated_at).split(".")[0]))
        data["training_start_at"] = predictor_record.training_start_at
        data["training_stop_at"] = predictor_record.training_stop_at
        data["predict"] = predictor_record.to_predict[0]
        data["update"] = predictor_record.update_status
        data["mindsdb_version"] = predictor_record.mindsdb_version
        data["name"] = predictor_record.name
        data["code"] = predictor_record.code
        data["problem_definition"] = predictor_record.learn_args
        data["fetch_data_query"] = predictor_record.fetch_data_query
        data["active"] = predictor_record.active
        data["status"] = predictor_record.status
        data["id"] = predictor_record.id
        data["version"] = predictor_record.version

        json_storage = get_json_storage(resource_id=predictor_record.id)
        data["json_ai"] = json_storage.get("json_ai")

        if data.get("accuracies", None) is not None:
            if len(data["accuracies"]) > 0:
                data["accuracy"] = float(np.mean(list(data["accuracies"].values())))
        return data

    def get_reduced_model_data(self, name: str = None, predictor_record=None, ml_handler_name="lightwood") -> dict:
        full_model_data = self.get_model_data(
            name=name, predictor_record=predictor_record, ml_handler_name=ml_handler_name
        )
        reduced_model_data = {}
        for k in [
            "id",
            "name",
            "version",
            "is_active",
            "predict",
            "status",
            "problem_definition",
            "current_phase",
            "accuracy",
            "data_source",
            "update",
            "active",
            "mindsdb_version",
            "error",
            "created_at",
            "fetch_data_query",
        ]:
            reduced_model_data[k] = full_model_data.get(k, None)

        reduced_model_data["training_time"] = None
        if full_model_data.get("training_start_at") is not None:
            if full_model_data.get("training_stop_at") is not None:
                reduced_model_data["training_time"] = full_model_data.get("training_stop_at") - full_model_data.get(
                    "training_start_at"
                )
            elif full_model_data.get("status") == "training":
                reduced_model_data["training_time"] = dt.datetime.now() - full_model_data.get("training_start_at")
            if reduced_model_data["training_time"] is not None:
                reduced_model_data["training_time"] = (
                    reduced_model_data["training_time"]
                    - dt.timedelta(microseconds=reduced_model_data["training_time"].microseconds)
                ).total_seconds()

        return reduced_model_data

    def describe_model(self, session, project_name, model_name, attribute, version=None):
        args = {"name": model_name, "version": version, "project_name": project_name, "except_absent": True}
        if version is not None:
            args["active"] = None

        model_record = get_model_record(**args)

        integration_record = db.Integration.query.get(model_record.integration_id)

        ml_handler_base = session.integration_controller.get_ml_handler(integration_record.name)

        return ml_handler_base.describe(model_record.id, attribute)

    def get_model(self, name, version=None, ml_handler_name=None, project_name=None):
        show_active = True if version is None else None
        model_record = get_model_record(
            active=show_active,
            version=version,
            name=name,
            ml_handler_name=ml_handler_name,
            except_absent=True,
            project_name=project_name,
        )
        data = self.get_reduced_model_data(predictor_record=model_record)
        integration_record = db.Integration.query.get(model_record.integration_id)
        if integration_record is not None:
            data["engine"] = integration_record.engine
            data["engine_name"] = integration_record.name
        return data

    def get_models(self, with_versions=False, ml_handler_name=None, integration_id=None, project_name=None):
        models = []
        show_active = True if with_versions is False else None
        for model_record in get_model_records(
            active=show_active,
            ml_handler_name=ml_handler_name,
            integration_id=integration_id,
            project_name=project_name,
        ):
            model_data = self.get_reduced_model_data(predictor_record=model_record)
            models.append(model_data)
        return models

    def delete_model(self, model_name: str, project_name: str = default_project, version=None):
        from mindsdb.interfaces.database.database import DatabaseController

        project_record = get_project_record(func.lower(project_name))
        if project_record is None:
            raise Exception(f"Project '{project_name}' does not exists")

        database_controller = DatabaseController()

        project = database_controller.get_project(project_name)

        if version is None:
            # Delete latest version
            predictors_records = get_model_records(
                name=model_name,
                project_id=project.id,
                active=None,
            )
        else:
            predictors_records = get_model_records(
                name=model_name,
                project_id=project.id,
                version=version,
            )
        if len(predictors_records) == 0:
            raise EntityNotExistsError("Model does not exist", model_name)

        is_cloud = self.config.get("cloud", False)
        if is_cloud:
            for predictor_record in predictors_records:
                model_data = self.get_model_data(predictor_record=predictor_record)
                if (
                    model_data.get("status") in ["generating", "training"]
                    and isinstance(model_data.get("created_at"), str) is True
                    and (dt.datetime.now() - parse_datetime(model_data.get("created_at"))) < dt.timedelta(hours=1)
                ):
                    raise Exception(
                        "You are unable to delete models currently in progress, please wait before trying again"
                    )

        for predictor_record in predictors_records:
            if is_cloud:
                predictor_record.deleted_at = dt.datetime.now()
            else:
                db.session.delete(predictor_record)
        db.session.commit()

        # region delete storages
        if len(predictors_records) > 1:
            ctx_dump = ctx.dump()
            with ThreadPool(min(len(predictors_records), 100)) as pool:
                pool.starmap(delete_model_storage, [(record.id, ctx_dump) for record in predictors_records])
        else:
            modelStorage = ModelStorage(predictors_records[0].id)
            modelStorage.delete()
        # endregion

    def rename_model(self, old_name, new_name):
        model_record = get_model_record(name=new_name)
        if model_record is None:
            raise Exception(f"Model with name '{new_name}' already exists")

        for model_record in get_model_records(name=old_name):
            model_record.name = new_name
        db.session.commit()

    @staticmethod
    def _get_data_integration_ref(statement, database_controller):
        # TODO use database_controller handler_controller internally
        data_integration_ref = None
        fetch_data_query = None
        if statement.integration_name is not None:
            fetch_data_query = statement.query_str
            integration_name = statement.integration_name.parts[0].lower()

            databases_meta = database_controller.get_dict()
            if integration_name not in databases_meta:
                raise EntityNotExistsError("Database does not exist", integration_name)
            data_integration_meta = databases_meta[integration_name]
            # TODO improve here. Suppose that it is view
            if data_integration_meta["type"] == "project":
                data_integration_ref = {"type": "project"}
            elif data_integration_meta["type"] == "system":
                data_integration_ref = {"type": "system"}
            else:
                data_integration_ref = {"type": "integration", "id": data_integration_meta["id"]}
        return data_integration_ref, fetch_data_query

    def prepare_create_statement(self, statement, database_controller):
        # extract data from Create model or Retrain statement and prepare it for using in crate and retrain functions
        project_name = statement.name.parts[0]
        model_name = statement.name.parts[1]

        sql_task = None
        if statement.task is not None:
            sql_task = statement.task.to_string()
        problem_definition = {"__mdb_sql_task": sql_task}
        if statement.targets is not None:
            problem_definition["target"] = statement.targets[0].parts[-1]

        data_integration_ref, fetch_data_query = self._get_data_integration_ref(statement, database_controller)

        label = None
        if statement.using is not None:
            label = statement.using.pop("tag", None)

            problem_definition["using"] = statement.using

        if statement.order_by is not None:
            problem_definition["timeseries_settings"] = {
                "is_timeseries": True,
                "order_by": getattr(statement, "order_by")[0].field.parts[-1],
            }
            for attr in ["horizon", "window"]:
                if getattr(statement, attr) is not None:
                    problem_definition["timeseries_settings"][attr] = getattr(statement, attr)

            if statement.group_by is not None:
                problem_definition["timeseries_settings"]["group_by"] = [col.parts[-1] for col in statement.group_by]

        join_learn_process = False
        if "join_learn_process" in problem_definition.get("using", {}):
            join_learn_process = problem_definition["using"]["join_learn_process"]
            del problem_definition["using"]["join_learn_process"]

        return dict(
            model_name=model_name,
            project_name=project_name,
            data_integration_ref=data_integration_ref,
            fetch_data_query=fetch_data_query,
            problem_definition=problem_definition,
            join_learn_process=join_learn_process,
            label=label,
        )

    def create_model(self, statement, ml_handler):
        params = self.prepare_create_statement(statement, ml_handler.database_controller)

        existing_projects_meta = ml_handler.database_controller.get_dict(filter_type="project", lowercase=False)
        if params["project_name"] not in existing_projects_meta:
            raise EntityNotExistsError("Project does not exist", params["project_name"])

        project = ml_handler.database_controller.get_project(name=params["project_name"], strict_case=True)
        project_tables = project.get_tables()
        if params["model_name"] in project_tables:
            raise EntityExistsError("Model already exists", f"{params['project_name']}.{params['model_name']}")
        predictor_record = ml_handler.learn(**params)

        return ModelController.get_model_info(predictor_record)

    def retrain_model(self, statement, ml_handler):
        # active setting
        set_active = True
        if statement.using is not None:
            set_active = statement.using.pop("active", True)
            if set_active in ("0", 0, None):
                set_active = False

        params = self.prepare_create_statement(statement, ml_handler.database_controller)

        base_predictor_record = get_model_record(
            name=params["model_name"], project_name=params["project_name"], active=True
        )

        model_name = params["model_name"]
        if base_predictor_record is None:
            raise Exception(f"Error: model '{model_name}' does not exist")

        if params["data_integration_ref"] is None:
            params["data_integration_ref"] = base_predictor_record.data_integration_ref
        if params["fetch_data_query"] is None:
            params["fetch_data_query"] = base_predictor_record.fetch_data_query

        problem_definition = base_predictor_record.learn_args.copy()
        problem_definition.update(params["problem_definition"])
        params["problem_definition"] = problem_definition

        params["is_retrain"] = True
        params["set_active"] = set_active
        predictor_record = ml_handler.learn(**params)

        return ModelController.get_model_info(predictor_record)

    def prepare_finetune_statement(self, statement, database_controller):
        project_name, model_name, model_version = resolve_model_identifier(statement.name)
        if project_name is None:
            project_name = default_project
        data_integration_ref, fetch_data_query = self._get_data_integration_ref(statement, database_controller)

        set_active = True
        if statement.using is not None:
            set_active = statement.using.pop("active", True)
            if set_active in ("0", 0, None):
                set_active = False

        label = None
        args = {}
        if statement.using is not None:
            label = statement.using.pop("tag", None)
            args = statement.using

        join_learn_process = args.pop("join_learn_process", False)

        base_predictor_record = get_model_record(
            name=model_name,
            project_name=project_name,
            version=model_version,
            active=True if model_version is None else None,
        )

        if data_integration_ref is None:
            data_integration_ref = base_predictor_record.data_integration_ref
        if fetch_data_query is None:
            fetch_data_query = base_predictor_record.fetch_data_query

        return dict(
            model_name=model_name,
            project_name=project_name,
            data_integration_ref=data_integration_ref,
            fetch_data_query=fetch_data_query,
            base_model_version=model_version,
            args=args,
            join_learn_process=join_learn_process,
            label=label,
            set_active=set_active,
        )

    @profiler.profile()
    def finetune_model(self, statement, ml_handler):
        params = self.prepare_finetune_statement(statement, ml_handler.database_controller)
        predictor_record = ml_handler.finetune(**params)
        return ModelController.get_model_info(predictor_record)

    def update_model(self, session, project_name: str, model_name: str, problem_definition, version=None):
        model_record = get_model_record(name=model_name, version=version, project_name=project_name, except_absent=True)
        integration_record = db.Integration.query.get(model_record.integration_id)

        ml_handler_base = session.integration_controller.get_ml_handler(integration_record.name)
        ml_handler_base.update(args=problem_definition, model_id=model_record.id)

        # update model record
        if "using" in problem_definition:
            learn_args = copy.deepcopy(model_record.learn_args)
            learn_args["using"].update(problem_definition["using"])
            model_record.learn_args = learn_args
            db.session.commit()

    @staticmethod
    def get_model_info(predictor_record):
        from mindsdb.interfaces.database.projects import ProjectController

        projects_controller = ProjectController()
        project = projects_controller.get(id=predictor_record.project_id)

        columns = [
            "NAME",
            "ENGINE",
            "PROJECT",
            "ACTIVE",
            "VERSION",
            "STATUS",
            "ACCURACY",
            "PREDICT",
            "UPDATE_STATUS",
            "MINDSDB_VERSION",
            "ERROR",
            "SELECT_DATA_QUERY",
            "TRAINING_OPTIONS",
            "TAG",
        ]

        project_name = project.name
        model = project.get_model_by_id(model_id=predictor_record.id)
        table_name = model["name"]
        table_meta = model["metadata"]
        record = [
            table_name,
            table_meta["engine"],
            project_name,
            table_meta["active"],
            table_meta["version"],
            table_meta["status"],
            table_meta["accuracy"],
            table_meta["predict"],
            table_meta["update_status"],
            table_meta["mindsdb_version"],
            table_meta["error"],
            table_meta["select_data_query"],
            str(table_meta["training_options"]),
            table_meta["label"],
        ]

        return pd.DataFrame([record], columns=columns)

    def set_model_active_version(self, project_name, model_name, version):
        model_record = get_model_record(name=model_name, project_name=project_name, version=version, active=None)

        if model_record is None:
            raise EntityNotExistsError(f"Model {model_name} with version {version} is not found in {project_name}")

        model_record.active = True

        # deactivate current active version
        model_records = db.Predictor.query.filter(
            db.Predictor.name == model_record.name,
            db.Predictor.project_id == model_record.project_id,
            db.Predictor.active == True,  # noqa
            db.Predictor.company_id == ctx.company_id,
            db.Predictor.id != model_record.id,
        )
        for p in model_records:
            p.active = False

        db.session.commit()

    def delete_model_version(self, project_name, model_name, version):
        model_record = get_model_record(name=model_name, project_name=project_name, version=version, active=None)
        if model_record is None:
            raise EntityNotExistsError(f"Model {model_name} with version {version} is not found in {project_name}")

        if model_record.active:
            raise Exception(f"Can't remove active version: {project_name}.{model_name}.{version}")

        is_cloud = self.config.get("cloud", False)
        if is_cloud:
            model_record.deleted_at = dt.datetime.now()
        else:
            db.session.delete(model_record)
        modelStorage = ModelStorage(model_record.id)
        modelStorage.delete()

        db.session.commit()



import shutil
import tempfile
from http import HTTPStatus
from typing import Dict
from pathlib import Path

from flask import request
from flask_restx import Resource

from mindsdb.api.http.utils import http_error
from mindsdb.api.http.namespaces.configs.databases import ns_conf
from mindsdb.api.mysql.mysql_proxy.classes.fake_mysql_proxy import FakeMysqlProxy
from mindsdb.api.executor.controllers.session_controller import SessionController
from mindsdb.api.executor.datahub.classes.tables_row import TablesRow
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE
from mindsdb.metrics.metrics import api_endpoint_metrics
from mindsdb_sql_parser import parse_sql, ParsingException
from mindsdb_sql_parser.ast import CreateTable, DropTables
from mindsdb.utilities.exception import EntityNotExistsError
from mindsdb.integrations.libs.response import HandlerStatusResponse
from mindsdb.utilities import log


logger = log.getLogger(__name__)


@ns_conf.route("/")
class DatabasesResource(Resource):
    @ns_conf.doc("list_databases")
    @api_endpoint_metrics("GET", "/databases")
    def get(self):
        """List all databases"""
        session = SessionController()
        return session.database_controller.get_list()

    @ns_conf.doc("create_database")
    @api_endpoint_metrics("POST", "/databases")
    def post(self):
        """Create a database"""
        if "database" not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST, "Wrong argument", 'Must provide "database" parameter in POST body'
            )
        check_connection = request.json.get("check_connection", False)
        session = SessionController()
        database = request.json["database"]
        parameters = {}
        if "name" not in database:
            return http_error(HTTPStatus.BAD_REQUEST, "Wrong argument", 'Missing "name" field for database')
        if "engine" not in database:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                "Wrong argument",
                'Missing "engine" field for database. If you want to create a project instead, use the /api/projects endpoint.',
            )
        if "parameters" in database:
            parameters = database["parameters"]
        name = database["name"]

        if session.database_controller.exists(name):
            return http_error(HTTPStatus.CONFLICT, "Name conflict", f"Database with name {name} already exists.")

        storage = None
        if check_connection:
            try:
                handler = session.integration_controller.create_tmp_handler(name, database["engine"], parameters)
                status = handler.check_connection()
            except ImportError as import_error:
                status = HandlerStatusResponse(success=False, error_message=str(import_error))

            if status.success is not True:
                if hasattr(status, "redirect_url") and isinstance(status, str):
                    return {
                        "status": "redirect_required",
                        "redirect_url": status.redirect_url,
                        "detail": status.error_message,
                    }, HTTPStatus.OK
                return {"status": "connection_error", "detail": status.error_message}, HTTPStatus.OK

            if status.copy_storage:
                storage = handler.handler_storage.export_files()

        new_integration_id = session.integration_controller.add(name, database["engine"], parameters)

        if storage:
            handler = session.integration_controller.get_data_handler(name, connect=False)
            handler.handler_storage.import_files(storage)

        new_integration = session.database_controller.get_integration(new_integration_id)
        return new_integration, HTTPStatus.CREATED


@ns_conf.route("/status")
class DatabasesStatusResource(Resource):
    @ns_conf.doc("check_database_connection_status")
    @api_endpoint_metrics("POST", "/databases/status")
    def post(self):
        """Check the connection parameters for a database"""
        data = {}
        if request.content_type == "application/json":
            data.update(request.json or {})
        elif request.content_type.startswith("multipart/form-data"):
            data.update(request.form or {})

        if "engine" not in data:
            return http_error(HTTPStatus.BAD_REQUEST, "Wrong argument", 'Missing "engine" field for database')

        engine = data["engine"]
        parameters = data
        del parameters["engine"]

        files = request.files
        temp_dir = None
        if files is not None and len(files) > 0:
            temp_dir = tempfile.mkdtemp(prefix="integration_files_")
            for key, file in files.items():
                temp_dir_path = Path(temp_dir)
                file_name = Path(file.filename)
                file_path = temp_dir_path.joinpath(file_name).resolve()
                if temp_dir_path not in file_path.parents:
                    raise Exception(f"Can not save file at path: {file_path}")
                file.save(file_path)
                parameters[key] = str(file_path)

        session = SessionController()

        try:
            handler = session.integration_controller.create_tmp_handler("test_connection", engine, parameters)
            status = handler.check_connection()
        except ImportError as import_error:
            status = HandlerStatusResponse(success=False, error_message=str(import_error))
        except Exception as unknown_error:
            status = HandlerStatusResponse(success=False, error_message=str(unknown_error))
        finally:
            if temp_dir is not None:
                shutil.rmtree(temp_dir)

        if not status.success:
            if hasattr(status, "redirect_url") and isinstance(status, str):
                return {
                    "status": "redirect_required",
                    "redirect_url": status.redirect_url,
                    "detail": status.error_message,
                }, HTTPStatus.OK
            return {"status": "connection_error", "detail": status.error_message}, HTTPStatus.OK

        return {
            "status": "success",
        }, HTTPStatus.OK


@ns_conf.route("/<database_name>")
class DatabaseResource(Resource):
    @ns_conf.doc("get_database")
    @api_endpoint_metrics("GET", "/databases/database")
    def get(self, database_name):
        """Gets a database by name"""
        session = SessionController()
        check_connection = request.args.get("check_connection", "false").lower() in ("1", "true")
        try:
            project = session.database_controller.get_project(database_name)
            result = {"name": database_name, "type": "project", "id": project.id, "engine": None}
            if check_connection:
                result["connection_status"] = {"success": True, "error_message": None}
        except (ValueError, EntityNotExistsError):
            integration = session.integration_controller.get(database_name)
            if integration is None:
                return http_error(
                    HTTPStatus.NOT_FOUND, "Database not found", f"Database with name {database_name} does not exist."
                )
            result = integration
            if check_connection:
                integration["connection_status"] = {"success": False, "error_message": None}
                try:
                    handler = session.integration_controller.get_data_handler(database_name)
                    status = handler.check_connection()
                    integration["connection_status"]["success"] = status.success
                    integration["connection_status"]["error_message"] = status.error_message
                except Exception as e:
                    integration["connection_status"]["success"] = False
                    integration["connection_status"]["error_message"] = str(e)

        return result

    @ns_conf.doc("update_database")
    @api_endpoint_metrics("PUT", "/databases/database")
    def put(self, database_name):
        """Updates or creates a database"""
        if "database" not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST, "Wrong argument", 'Must provide "database" parameter in POST body'
            )

        session = SessionController()
        parameters = {}
        database = request.json["database"]
        check_connection = request.json.get("check_connection", False)

        if "parameters" in database:
            parameters = database["parameters"]
        if not session.database_controller.exists(database_name):
            # Create.
            if "engine" not in database:
                return http_error(
                    HTTPStatus.BAD_REQUEST,
                    "Wrong argument",
                    'Missing "engine" field for new database. '
                    "If you want to create a project instead, use the POST /api/projects endpoint.",
                )
            new_integration_id = session.integration_controller.add(database_name, database["engine"], parameters)
            new_integration = session.database_controller.get_integration(new_integration_id)
            return new_integration, HTTPStatus.CREATED

        try:
            session.integration_controller.modify(database_name, parameters, check_connection=check_connection)
        except Exception as e:
            status = HandlerStatusResponse(success=False, error_message=str(e))
            return http_error(HTTPStatus.BAD_REQUEST, "Connection error", status.error_message or "Connection error")

        return session.integration_controller.get(database_name)

    @ns_conf.doc("delete_database")
    @api_endpoint_metrics("DELETE", "/databases/database")
    def delete(self, database_name):
        """Deletes a database by name"""
        session = SessionController()
        if not session.database_controller.exists(database_name):
            return http_error(
                HTTPStatus.NOT_FOUND, "Database not found", f"Database with name {database_name} does not exist."
            )
        try:
            session.database_controller.delete(database_name)
        except Exception as e:
            logger.debug(f"Error while deleting database '{database_name}'", exc_info=True)
            return http_error(
                HTTPStatus.BAD_REQUEST,
                "Error",
                f"Cannot delete database {database_name}. "
                + "This is most likely a system database, a permanent integration, or an ML engine with active models. "
                + f"Full error: {e}. "
                + "Please check the name and try again.",
            )
        return "", HTTPStatus.NO_CONTENT


def _tables_row_to_obj(table_row: TablesRow) -> Dict:
    type = table_row.TABLE_TYPE.lower()
    if table_row.TABLE_TYPE == "BASE TABLE":
        type = "data"
    return {"name": table_row.TABLE_NAME, "type": type}


@ns_conf.route("/<database_name>/tables")
class TablesList(Resource):
    @ns_conf.doc("list_tables")
    @api_endpoint_metrics("GET", "/databases/database/tables")
    def get(self, database_name):
        """Get all tables in a database"""
        session = SessionController()
        datanode = session.datahub.get(database_name)
        all_tables = datanode.get_tables()
        table_objs = [_tables_row_to_obj(t) for t in all_tables]
        return table_objs

    @ns_conf.doc("create_table")
    @api_endpoint_metrics("POST", "/databases/database/tables")
    def post(self, database_name):
        """Creates a table in a database"""
        if "table" not in request.json:
            return http_error(HTTPStatus.BAD_REQUEST, "Wrong argument", 'Must provide "table" parameter in POST body')
        table = request.json["table"]
        if "name" not in table:
            return http_error(HTTPStatus.BAD_REQUEST, "Wrong argument", 'Missing "name" field for table')
        if "select" not in table:
            return http_error(HTTPStatus.BAD_REQUEST, "Wrong argument", 'Missing "select" field for table')
        table_name = table["name"]
        select_query = table["select"]
        replace = False
        if "replace" in table:
            replace = table["replace"]

        session = SessionController()
        try:
            session.database_controller.get_project(database_name)
            error_message = (
                f"Database {database_name} is a project. "
                + f"If you want to create a model or view, use the projects/{database_name}/models/{table_name} or "
                + f"projects/{database_name}/views/{table_name} endpoints instead."
            )
            return http_error(HTTPStatus.BAD_REQUEST, "Error", error_message)
        except EntityNotExistsError:
            # Only support creating tables from integrations.
            pass

        datanode = session.datahub.get(database_name)
        if datanode is None:
            return http_error(
                HTTPStatus.NOT_FOUND, "Database not exists", f"Database with name {database_name} does not exist"
            )
        all_tables = datanode.get_tables()
        for t in all_tables:
            if t.TABLE_NAME == table_name and not replace:
                return http_error(HTTPStatus.CONFLICT, "Name conflict", f"Table with name {table_name} already exists")

        try:
            select_ast = parse_sql(select_query)
        except ParsingException:
            return http_error(HTTPStatus.BAD_REQUEST, "Error", f"Could not parse select statement {select_query}")

        create_ast = CreateTable(f"{database_name}.{table_name}", from_select=select_ast, is_replace=replace)

        mysql_proxy = FakeMysqlProxy()

        try:
            mysql_proxy.process_query(create_ast.get_string())
        except Exception as e:
            return http_error(HTTPStatus.BAD_REQUEST, "Error", str(e))

        all_tables = datanode.get_tables()
        try:
            matching_table = next(t for t in all_tables if t.TABLE_NAME == table_name)
            return _tables_row_to_obj(matching_table), HTTPStatus.CREATED
        except StopIteration:
            return http_error(
                HTTPStatus.INTERNAL_SERVER_ERROR, "Error", f"Table with name {table_name} could not be created"
            )


@ns_conf.route("/<database_name>/tables/<table_name>")
@ns_conf.param("database_name", "Name of the database")
@ns_conf.param("table_name", "Name of the table")
class TableResource(Resource):
    @ns_conf.doc("get_table")
    @api_endpoint_metrics("GET", "/databases/database/tables/table")
    def get(self, database_name, table_name):
        session = SessionController()
        datanode = session.datahub.get(database_name)
        all_tables = datanode.get_tables()
        try:
            matching_table = next(t for t in all_tables if t.TABLE_NAME == table_name)
            return _tables_row_to_obj(matching_table)
        except StopIteration:
            return http_error(HTTPStatus.NOT_FOUND, "Table not found", f"Table with name {table_name} not found")

    @ns_conf.doc("drop_table")
    @api_endpoint_metrics("DELETE", "/databases/database/tables/table")
    def delete(self, database_name, table_name):
        session = SessionController()
        try:
            session.database_controller.get_project(database_name)
            error_message = (
                f"Database {database_name} is a project. "
                + f"If you want to delete a model or view, use the projects/{database_name}/models/{table_name} or "
                + f"projects/{database_name}/views/{table_name} endpoints instead."
            )
            return http_error(HTTPStatus.BAD_REQUEST, "Error", error_message)
        except EntityNotExistsError:
            # Only support dropping tables from integrations.
            pass

        datanode = session.datahub.get(database_name)
        if datanode is None:
            return http_error(
                HTTPStatus.NOT_FOUND, "Database not found", f"Database with name {database_name} not found"
            )
        all_tables = datanode.get_tables()
        try:
            next(t for t in all_tables if t.TABLE_NAME == table_name)
        except StopIteration:
            return http_error(HTTPStatus.NOT_FOUND, "Table not found", f"Table with name {table_name} not found")

        drop_ast = DropTables(tables=[table_name], if_exists=True)

        try:
            integration_handler = session.integration_controller.get_data_handler(database_name)
        except Exception:
            return http_error(
                HTTPStatus.INTERNAL_SERVER_ERROR, "Error", f"Could not get database handler for {database_name}"
            )
        try:
            result = integration_handler.query(drop_ast)
        except NotImplementedError:
            return http_error(
                HTTPStatus.BAD_REQUEST, "Error", f"Database {database_name} does not support dropping tables."
            )
        if result.type == RESPONSE_TYPE.ERROR:
            return http_error(HTTPStatus.BAD_REQUEST, "Error", result.error_message)
        return "", HTTPStatus.NO_CONTENT



import unittest
from unittest.mock import patch, MagicMock
from mindsdb.integrations.handlers.bigcommerce_handler.bigcommerce_handler import BigCommerceHandler
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE
from mindsdb_sql_parser import ast


class BigCommerceHandlerTest(unittest.TestCase):
    def setUp(self):
        self.handler = BigCommerceHandler(
            "test_bigcommerce_handler",
            connection_data={
                "api_base": "https://api.bigcommerce.com/stores/test-store/v3/",
                "access_token": "mock_access_token",
            },
        )
        self.patcher = patch(
            "mindsdb.integrations.handlers.bigcommerce_handler.bigcommerce_handler.BigCommerceAPIClient"
        )
        self.mock_client = self.patcher.start()
        self.mock_client_instance = MagicMock()
        self.mock_client.return_value = self.mock_client_instance

    def tearDown(self):
        self.patcher.stop()

    def test_check_connection_success(self):
        """Test successful connection to BigCommerce API"""
        self.mock_client_instance.get_products.return_value = [{"id": 1, "name": "Test Product"}]
        response = self.handler.check_connection()
        self.assertTrue(response.success)
        self.mock_client_instance.get_products.assert_called_once_with(limit=1)

    def test_check_connection_failure(self):
        """Test failed connection to BigCommerce API"""
        self.mock_client_instance.get_products.side_effect = Exception("Connection failed")
        response = self.handler.check_connection()
        self.assertFalse(response.success)
        self.assertIsNotNone(response.error_message)

    def test_get_tables(self):
        """Test retrieving list of tables"""
        result = self.handler.get_tables()
        self.assertIsNotNone(result)
        self.assertEqual(result.type, RESPONSE_TYPE.TABLE)
        # Check that all expected tables are present
        table_names = [row[0] for row in result.data_frame.values]
        expected_tables = [
            "orders",
            "products",
            "customers",
            "categories",
            "pickups",
            "promotions",
            "wishlists",
            "segments",
            "brands",
        ]
        for table in expected_tables:
            self.assertIn(table, table_names)

    def test_query_products(self):
        """Test querying products table"""
        mock_products = [
            {
                "id": 1,
                "name": "Product 1",
                "type": "physical",
                "sku": "SKU-001",
                "price": "29.99",
                "inventory_level": 100,
            },
            {
                "id": 2,
                "name": "Product 2",
                "type": "digital",
                "sku": "SKU-002",
                "price": "49.99",
                "inventory_level": 50,
            },
        ]
        self.mock_client_instance.get_products.return_value = mock_products

        query = ast.Select(targets=[ast.Star()], from_table=ast.Identifier("products"), limit=ast.Constant(10))

        self.handler.connect = MagicMock(return_value=self.mock_client_instance)
        result = self.handler.query(query)

        self.assertEqual(result.type, RESPONSE_TYPE.TABLE)
        self.assertEqual(len(result.data_frame), 2)

    def test_query_orders(self):
        """Test querying orders table"""
        mock_orders = [
            {"id": 100, "customer_id": 1, "status": "completed", "total_inc_tax": 99.99, "date_created": "2024-01-15"},
            {"id": 101, "customer_id": 2, "status": "pending", "total_inc_tax": 149.99, "date_created": "2024-01-16"},
        ]
        self.mock_client_instance.get_orders.return_value = mock_orders

        query = ast.Select(targets=[ast.Star()], from_table=ast.Identifier("orders"), limit=ast.Constant(10))

        self.handler.connect = MagicMock(return_value=self.mock_client_instance)
        result = self.handler.query(query)

        self.assertEqual(result.type, RESPONSE_TYPE.TABLE)
        self.assertEqual(len(result.data_frame), 2)

    def test_query_customers(self):
        """Test querying customers table"""
        mock_customers = [
            {
                "id": 1,
                "email": "customer1@example.com",
                "first_name": "John",
                "last_name": "Doe",
                "customer_group_id": 0,
            },
            {
                "id": 2,
                "email": "customer2@example.com",
                "first_name": "Jane",
                "last_name": "Smith",
                "customer_group_id": 1,
            },
        ]
        self.mock_client_instance.get_customers.return_value = mock_customers

        query = ast.Select(targets=[ast.Star()], from_table=ast.Identifier("customers"), limit=ast.Constant(10))

        self.handler.connect = MagicMock(return_value=self.mock_client_instance)
        result = self.handler.query(query)

        self.assertEqual(result.type, RESPONSE_TYPE.TABLE)
        self.assertEqual(len(result.data_frame), 2)

    def test_query_with_filter(self):
        """Test querying with WHERE clause"""
        mock_products = [{"id": 1, "name": "Filtered Product", "price": 29.99}]
        self.mock_client_instance.get_products.return_value = mock_products

        query = ast.Select(
            targets=[ast.Star()],
            from_table=ast.Identifier("products"),
            where=ast.BinaryOperation("=", args=[ast.Identifier("id"), ast.Constant(1)]),
            limit=ast.Constant(10),
        )

        self.handler.connect = MagicMock(return_value=self.mock_client_instance)
        result = self.handler.query(query)

        self.assertEqual(result.type, RESPONSE_TYPE.TABLE)
        self.assertGreaterEqual(len(result.data_frame), 0)

    def test_connect_missing_parameters(self):
        """Test connection with missing required parameters"""
        handler = BigCommerceHandler(
            "test_handler",
            connection_data={
                "api_base": "https://api.bigcommerce.com/stores/test/"
                # Missing access_token
            },
        )

        with self.assertRaises(ValueError) as context:
            handler.connect()

        self.assertIn("Required parameters", str(context.exception))

    def test_meta_columns(self):
        """Test retrieving metadata columns"""
        meta_columns = self.handler.meta_get_columns()
        self.assertTrue(len(meta_columns.data_frame) > 0)


if __name__ == "__main__":
    unittest.main()



import os
import shutil
import tarfile
import tempfile
import zipfile
from pathlib import Path
from urllib.parse import urlparse

import multipart
import requests
from flask import current_app as ca
from flask import request
from flask_restx import Resource

from mindsdb.api.http.namespaces.configs.files import ns_conf
from mindsdb.api.http.utils import http_error
from mindsdb.metrics.metrics import api_endpoint_metrics
from mindsdb.utilities.config import config
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities import log
from mindsdb.utilities.security import is_private_url, clear_filename, validate_urls
from mindsdb.utilities.fs import safe_extract
from mindsdb.integrations.utilities.files.file_reader import FileProcessingError

logger = log.getLogger(__name__)
MAX_FILE_SIZE = 1024 * 1024 * 100  # 100Mb


@ns_conf.route("/")
class FilesList(Resource):
    @ns_conf.doc("get_files_list")
    @api_endpoint_metrics("GET", "/files")
    def get(self):
        """List all files"""
        return ca.file_controller.get_files()


@ns_conf.route("/<name>")
@ns_conf.param("name", "MindsDB's name for file")
class File(Resource):
    @ns_conf.doc("put_file")
    @api_endpoint_metrics("PUT", "/files/file")
    def put(self, name: str):
        """add new file as table

        The table name is <name> path paramether

        Data is provided as json or form data. File can be provided with FormData or via URL.

        If file is in FormData, then the form contain:
            - source_type (str) - 'file'
            - file (binary) - the file itself
            - original_file_name (str, optional) - the name with which the file will be saved

        If file should be uploaded from URL:
            - source_type (str) - 'url'
            - source (str) - the URL
            - original_file_name (str, optional) - the name with which the file will be saved
        """

        data = {}
        mindsdb_file_name = name.lower()

        def on_field(field):
            name = field.field_name.decode()
            value = field.value.decode()
            data[name] = value

        file_object = None

        def on_file(file):
            nonlocal file_object
            file_name = file.file_name.decode()
            data["file"] = file_name
            if Path(file_name).name != file_name:
                raise ValueError(f"Wrong file name: {file_name}")
            file_object = file.file_object

        temp_dir_path = tempfile.mkdtemp(prefix="mindsdb_file_")

        if request.headers["Content-Type"].startswith("multipart/form-data"):
            parser = multipart.create_form_parser(
                headers=request.headers,
                on_field=on_field,
                on_file=on_file,
                config={
                    "UPLOAD_DIR": temp_dir_path.encode(),  # bytes required
                    "UPLOAD_KEEP_FILENAME": False,
                    "UPLOAD_KEEP_EXTENSIONS": True,
                    "UPLOAD_DELETE_TMP": False,
                    "MAX_MEMORY_FILE_SIZE": 0,
                },
            )

            while True:
                chunk = request.stream.read(8192)
                if not chunk:
                    break
                parser.write(chunk)
            parser.finalize()
            parser.close()

            if file_object is not None:
                if not file_object.closed:
                    try:
                        file_object.flush()
                    except (AttributeError, ValueError, OSError):
                        logger.debug("Failed to flush file_object before closing.", exc_info=True)
                    file_object.close()
                Path(file_object.name).rename(Path(file_object.name).parent / data["file"])
                file_object = None
        else:
            data = request.json

        existing_file_names = ca.file_controller.get_files_names(lower=True)
        if mindsdb_file_name.lower() in existing_file_names:
            return http_error(
                400,
                "File already exists",
                f"File with name '{mindsdb_file_name}' already exists",
            )

        source_type = data.get("source_type", "file")
        if source_type not in ("file", "url"):
            return http_error(
                400,
                "Wrong file source type",
                f'Only "file" and "url" supported as file source, got "{source_type}"',
            )

        if source_type == "url":
            url_file_upload_enabled = config["url_file_upload"]["enabled"]
            if url_file_upload_enabled is False:
                return http_error(400, "URL file upload is disabled.", "URL file upload is disabled.")

            if "file" in data:
                return http_error(
                    400,
                    "Fields conflict",
                    'URL source type can not be used together with "file" field.',
                )
            if "source" not in data:
                return http_error(
                    400,
                    "Missed file source",
                    'If the file\'s source type is URL, the "source" field should be specified.',
                )
            url = data["source"]
            try:
                url = urlparse(url)
                if not (url.scheme and url.netloc):
                    raise ValueError()
                url = url.geturl()
            except Exception:
                return http_error(
                    400,
                    "Invalid URL",
                    f"The URL is not valid: {data['source']}",
                )

            allowed_origins = config["url_file_upload"]["allowed_origins"]
            disallowed_origins = config["url_file_upload"]["disallowed_origins"]

            if validate_urls(url, allowed_origins, disallowed_origins) is False:
                return http_error(
                    400,
                    "Invalid URL",
                    "URL is not allowed for security reasons. Allowed hosts are: "
                    f"{', '.join(allowed_origins) if allowed_origins else 'not specified'}.",
                )

            data["file"] = clear_filename(mindsdb_file_name)
            if config.is_cloud:
                if is_private_url(url):
                    return http_error(400, f"URL is private: {url}")

                if ctx.user_class != 1:
                    info = requests.head(url, timeout=30)
                    file_size = info.headers.get("Content-Length")
                    try:
                        file_size = int(file_size)
                    except Exception:
                        pass

                    if file_size is None:
                        return http_error(
                            400,
                            "Error getting file info",
                            "Can't determine remote file size",
                        )
                    if file_size > MAX_FILE_SIZE:
                        return http_error(
                            400,
                            "File is too big",
                            f"Upload limit for file is {MAX_FILE_SIZE >> 20} MB",
                        )
            with requests.get(url, stream=True) as r:
                if r.status_code != 200:
                    return http_error(400, "Error getting file", f"Got status code: {r.status_code}")
                file_path = os.path.join(temp_dir_path, data["file"])
                with open(file_path, "wb") as f:
                    for chunk in r.iter_content(chunk_size=8192):
                        f.write(chunk)

        if "file" not in data:
            return http_error(
                400,
                "File field is missed",
                'The "field" field is missed in the form',
            )

        original_file_name = clear_filename(data.get("original_file_name"))

        file_path = os.path.join(temp_dir_path, data["file"])
        lp = file_path.lower()
        if lp.endswith((".zip", ".tar.gz")):
            if lp.endswith(".zip"):
                with zipfile.ZipFile(file_path) as f:
                    f.extractall(temp_dir_path)
            elif lp.endswith(".tar.gz"):
                with tarfile.open(file_path) as f:
                    safe_extract(f, temp_dir_path)
            os.remove(file_path)
            files = os.listdir(temp_dir_path)
            if len(files) != 1:
                os.rmdir(temp_dir_path)
                return http_error(400, "Wrong content.", "Archive must contain only one data file.")
            file_path = os.path.join(temp_dir_path, files[0])
            mindsdb_file_name = files[0]
            if not os.path.isfile(file_path):
                os.rmdir(temp_dir_path)
                return http_error(400, "Wrong content.", "Archive must contain data file in root.")

        try:
            if not Path(mindsdb_file_name).suffix == "":
                return http_error(400, "Error", "File name cannot contain extension.")
            ca.file_controller.save_file(mindsdb_file_name, file_path, file_name=original_file_name)
        except FileProcessingError as e:
            return http_error(400, "Error", str(e))
        except Exception as e:
            return http_error(500, "Error", str(e))
        finally:
            shutil.rmtree(temp_dir_path, ignore_errors=True)

        return "", 200

    @ns_conf.doc("delete_file")
    @api_endpoint_metrics("DELETE", "/files/file")
    def delete(self, name: str):
        """delete file"""

        try:
            ca.file_controller.delete_file(name)
        except FileNotFoundError:
            logger.exception(f"Error when deleting file '{name}'")
            return http_error(
                400,
                "Error deleting file",
                f"There was an error while trying to delete file with name '{name}'",
            )
        except Exception as e:
            logger.error(e)
            return http_error(
                500,
                "Error occured while deleting file",
                f"There was an error while trying to delete file with name '{name}'",
            )
        return "", 200



from typing import Optional
from pydantic import BaseModel, AnyUrl, TypeAdapter, model_validator, field_validator, ConfigDict
from urllib.parse import urlparse


_ANY_URL_ADAPTER = TypeAdapter(AnyUrl)


class ConnectionConfig(BaseModel):
    """
    MySQL connection configuration with validation.

    Supports two connection methods:
    1. URL-based: mysql://user:password@host:port/database
    2. Parameter-based: individual host, port, user, password, database params
    """

    url: Optional[AnyUrl] = None
    host: Optional[str] = None
    port: int = 3306
    user: Optional[str] = None
    password: Optional[str] = None
    database: Optional[str] = None

    @field_validator("port")
    @classmethod
    def validate_port(cls, v: int) -> int:
        """Validate that port is within valid range."""
        if v < 1 or v > 65535:
            raise ValueError(f"Port must be between 1 and 65535, got {v}")
        return v

    @field_validator("url", mode="before")
    @classmethod
    def validate_url(cls, v: Optional[str]) -> Optional[AnyUrl]:
        """Validate URL using AnyUrl as a fallback option for MySQL DSN parsing."""
        if v is None or isinstance(v, AnyUrl):
            return v
        try:
            return _ANY_URL_ADAPTER.validate_python(v)
        except ValueError as exc:
            raise ValueError(f"Invalid MySQL connection URL: {v}") from exc

    @field_validator("host")
    @classmethod
    def validate_host(cls, v: Optional[str]) -> Optional[str]:
        """Validate that host is not empty if provided."""
        if v is not None and not v.strip():
            raise ValueError("Host cannot be empty string")
        return v

    @field_validator("database")
    @classmethod
    def validate_database(cls, v: Optional[str]) -> Optional[str]:
        """Validate that database name is not empty if provided."""
        if v is not None and not v.strip():
            raise ValueError("Database name cannot be empty string")
        return v

    @model_validator(mode="before")
    @classmethod
    def check_db_params(cls, values):
        """Ensures either URL is provided or all individual parameters are provided."""
        url = values.get("url")
        host = values.get("host")
        user = values.get("user")
        password = values.get("password")
        database = values.get("database")

        if not url and not (host and user and password and database):
            missing_params = []
            if not host:
                missing_params.append("host")
            if not user:
                missing_params.append("user")
            if not password:
                missing_params.append("password")
            if not database:
                missing_params.append("database")

            raise ValueError(
                f"Either a valid URL or all required parameters must be provided. Missing: {', '.join(missing_params)}"
            )

        if url:
            # Parse URL and extract connection parameters
            try:
                parsed = urlparse(str(url))

                # Extract parameters from URL
                values["host"] = parsed.hostname or host
                values["port"] = parsed.port if parsed.port is not None else values.get("port", 3306)
                values["user"] = parsed.username or user
                values["password"] = parsed.password or password
                values["database"] = parsed.path[1:] if parsed.path and len(parsed.path) > 1 else database

                # Validate extracted parameters
                if not values["host"]:
                    raise ValueError("URL must contain a hostname")
                if not values["user"]:
                    raise ValueError("URL must contain a username")
                if not values["database"]:
                    raise ValueError("URL must contain a database name in the path")

            except Exception as e:
                raise ValueError(f"Invalid MySQL connection URL: {str(e)}")

            # mysql connector raises error if url is provided
            values.pop("url", None)

            return values

        # Validate individual parameters
        if not url:
            for param in ["host", "user", "password", "database"]:
                if not values.get(param):
                    raise ValueError(f"'{param}' is required when URL is not provided")

        return values

    model_config = ConfigDict(str_min_length=1, str_strip_whitespace=True, validate_assignment=True)
